import pytest
import pandas as pd
import numpy as np
import os
from jinja2 import Template
import json
import itertools
import nway.utils as utils
import nway.pairwise_matching as pairwise

TEST_FILE_DIR = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        'test_files')


@pytest.fixture(scope='function')
def input_file(tmpdir):
    thistest = os.path.join(TEST_FILE_DIR, 'test0')
    myinput = os.path.join(thistest, 'input.json')
    with open(myinput, 'r') as f:
        template = Template(json.dumps(json.load(f)))
    output_dir = str(tmpdir.mkdir("pairwise_test"))
    rendered = json.loads(
            template.render(
                output_dir=output_dir,
                test_files_dir=str(thistest)))
    input_json = os.path.join(output_dir, 'input.json')
    with open(input_json, 'w') as f:
        json.dump(rendered, f, indent=2)
    yield input_json


def read_pairs_from_simple(cost, val):
    inds = np.argwhere(np.array(cost) == val)
    pairs = []
    for i in inds:
        pairs.append([
            cost.index[i[0]],
            cost.columns[i[1]]])
    return pairs


def shuffle_dataframe(df):
    cshuff = list(df.columns)
    np.random.shuffle(cshuff)
    df = df[cshuff]
    df = df.sample(frac=1)
    return df


@pytest.mark.parametrize('shuffle, niter', zip([True, False], [20, 1]))
@pytest.mark.parametrize('solver', ['Hungarian', 'Blossom'])
def test_solvers_easy(solver, shuffle, niter):
    nrow = 50
    ncol = 100
    rows = ['r_%d' % i for i in range(nrow)]
    cols = ['c_%d' % i for i in range(ncol)]
    high_cost = 1000.0
    low_cost = 1.0
    cost_arr = np.ones((nrow, ncol)) * high_cost
    for i in range(nrow):
        cost_arr[i, i] = low_cost

    cost = pd.DataFrame(cost_arr, rows, cols)

    # ground truth for this test
    validation_pairs = read_pairs_from_simple(cost, low_cost)
    validation_set = set([tuple(i) for i in validation_pairs])

    test_sets = []
    for i in range(niter):
        # result should not be order-dependent
        if shuffle:
            cost = shuffle_dataframe(cost)

        pairs = pairwise.gen_assignment_pairs(cost, solver)

        test_sets.append(set([tuple(i) for i in pairs]))

    for test_set in test_sets:
        assert validation_set == test_set


@pytest.mark.parametrize('solver', ['Hungarian', 'Blossom'])
def test_real_cost_data(input_file, tmpdir, solver):
    # Hungarian method only stable under permutations at the ~80% level
    # Blossom stable at the > 95% level
    assert os.path.isfile(input_file)
    with open(input_file, 'r') as f:
        j = json.load(f)

    output_dir = str(tmpdir.mkdir("real_cost"))

    fixed = utils.create_nice_mask(
            j['experiment_containers']['ophys_experiments'][0],
            output_dir)
    moving = utils.create_nice_mask(
            j['experiment_containers']['ophys_experiments'][1],
            output_dir)

    # NOTE specifying solver here only matters for code coverage
    # of input schema, as the cost matrix is used afterwards
    # to calculate pairs.
    pair_args = {
            'fixed': fixed,
            'moving': moving,
            'output_directory': output_dir,
            'assignment_solver': solver,
            'edge_buffer': 0,
            'output_json': os.path.join(
                    output_dir,
                    "{}_to_{}_output.json".format(moving['id'], fixed['id'])),
            'log_level': "DEBUG"
            }

    pm = pairwise.PairwiseMatching(input_data=pair_args, args=[])
    pm.run()

    cost = pm.cost_matrix

    def iou(a, b):
        i = len(a.intersection(b))
        u = len(a.union(b))
        return float(i) / u

    test_sets = []
    for i in range(100):
        cost = shuffle_dataframe(cost)

        pairs = pairwise.gen_assignment_pairs(cost, solver)
        test_sets.append(set([tuple(i) for i in pairs]))

    combos = np.array([
        iou(a, b) for a, b in itertools.combinations(test_sets, 2)])

    if 'Hungarian' in solver:
        assert combos.mean() > 0.75
        assert combos.mean() < 0.85
    else:
        assert combos.mean() > 0.95


def test_gen_matching_table():
    nrow = 10
    ncol = 15
    rows = ['r_%d' % i for i in range(nrow)]
    cols = ['c_%d' % i for i in range(ncol)]
    high_cost = 1000.0
    low_cost = 1.0
    cost_arr = np.ones((nrow, ncol)) * high_cost
    for i in range(nrow):
        cost_arr[i, i] = low_cost
    cost = pd.DataFrame(cost_arr, rows, cols)

    validation_pairs = read_pairs_from_simple(cost, low_cost)

    distance = pd.DataFrame(np.ones(cost_arr.shape) * 1000, rows, cols)
    iou = pd.DataFrame(np.random.randn(*cost_arr.shape), rows, cols)

    for pair in validation_pairs:
        distance[pair[1]][pair[0]] = 1.234
        iou[pair[1]][pair[0]] = 5.678

    # this pair will not be in matches, but in rejected
    # it is below the distance threshold, but it does not
    # show up in the validation_pairs list
    rej_pair = ['r_9', 'c_11']
    distance[rej_pair[1]][rej_pair[0]] = 1.234

    matches, rejected = pairwise.gen_matching_table(
            cost, validation_pairs, distance, iou, 10.0)

    valsets = set([tuple(i) for i in validation_pairs])
    msets = set([(m['fixed'], m['moving']) for m in matches])
    assert valsets == msets
    mdists = np.array([m['distance'] for m in matches])
    assert np.all(np.isclose(mdists, 1.234))
    mious = np.array([m['iou'] for m in matches])
    assert np.all(np.isclose(mious, 5.678))

    rvalsets = set([tuple(i) for i in [rej_pair]])
    rmsets = set([(m['fixed'], m['moving']) for m in rejected])
    assert rvalsets == rmsets
    assert np.isclose(rejected[0]['distance'], 1.234)


def test_region_properties():
    mask = np.zeros((3, 100, 100)).astype('uint16')
    mask[0, 40:50, 40:50] = 1
    i1 = np.argwhere(mask[0] == 1)
    mask[1, 60:65, 30:35] = 2
    i2 = np.argwhere(mask[1] == 2)
    mdict = {
            'mask_dict': {
                "1": 123456,
                "2": 789012,
                }
            }

    prop = pairwise.region_properties(mask, mdict)

    assert prop['centers'].shape == (2, 2)
    ic1 = i1.mean(axis=0)[::-1]
    ic2 = i2.mean(axis=0)[::-1]
    assert np.all(np.isclose(
        prop['centers'],
        np.array([ic1, ic2])))


def test_calc_distance_iou():
    mask1 = np.zeros((3, 100, 100)).astype('uint16')
    mask1[0, 40:50, 40:50] = 1
    mask1[1, 60:65, 30:35] = 2
    dict1 = {
            'mask_dict': {
                "1": 123456,
                "2": 789012,
                }
            }
    mask2 = np.zeros((3, 100, 100)).astype('uint16')
    mask2[0, 38:50, 45:50] = 1
    mask2[1, 63:70, 30:35] = 2
    dict2 = {
            'mask_dict': {
                "1": 987654,
                "2": 321987,
                "3": 100000
                }
            }

    distance, iou = pairwise.calculate_distance_and_iou(
            mask1,
            mask2,
            dict1,
            dict2)

    assert set(distance.columns.tolist()) == set(dict2['mask_dict'].values())
    assert set(distance.index.tolist()) == set(dict1['mask_dict'].values())
    assert set(iou.columns.tolist()) == set(dict2['mask_dict'].values())
    assert set(iou.index.tolist()) == set(dict1['mask_dict'].values())


def test_calculate_cost_matrix():
    nrow = 10
    ncol = 15
    rows = ['r_%d' % i for i in range(nrow)]
    cols = ['c_%d' % i for i in range(ncol)]
    darr = np.random.rand(nrow, ncol) * 100
    distance = pd.DataFrame(darr, rows, cols)
    iarr = np.ones((nrow, ncol)) * 0.5
    iou = pd.DataFrame(iarr, rows, cols)

    cost = pairwise.calculate_cost_matrix(distance, iou, 10)
    assert np.all(cost.columns.tolist() == distance.columns.tolist())
    assert np.all(cost.index.tolist() == distance.index.tolist())

    carr = np.array(cost)

    ind = darr > 10
    assert np.all(carr[ind] == 999.5)
    ind = darr <= 10
    assert np.all(carr[ind] == (darr[ind]/10 + 0.5))



import pytest
import nway.utils as utils
import numpy as np
from jinja2 import Template
import os
import json
import PIL.Image

TEST_FILE_DIR = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        'test_files')


@pytest.fixture(scope='function')
def experiments(tmpdir):
    thistest = os.path.join(TEST_FILE_DIR, 'test0')
    myinput = os.path.join(thistest, 'input.json')
    with open(myinput, 'r') as f:
        template = Template(json.dumps(json.load(f)))
    output_dir = str(tmpdir.mkdir("nway_test"))
    rendered = json.loads(
            template.render(
                output_dir=output_dir,
                test_files_dir=str(thistest)))
    yield rendered['experiment_containers']['ophys_experiments']


@pytest.mark.parametrize("suptitle", [None, "stitle"])
def test_write_pair_images(suptitle, tmpdir):
    x = np.random.randint(0, 255, size=(512, 512)).astype('uint8')
    fname = tmpdir / "test.png"
    fname = utils.write_pair_images(fname, ['a', 'b'], [x, x],
                                    suptitle=suptitle)
    assert fname.exists()


def test_write_pair_images_exception(tmpdir):
    x = np.random.randint(0, 255, size=(512, 512)).astype('uint8')
    fname = tmpdir / "test.png"
    with pytest.raises(ValueError, match=r"length of titles.*"):
        fname = utils.write_pair_images(fname, ['a', 'b', 'c'], [x, x])


def test_summarize_registration_success(tmpdir):
    pair_outputs = [
            {
                'fixed_experiment': 123,
                'moving_experiment': 234,
                'transform': {'best_registration': ['Identity']}},
            {
                'fixed_experiment': 234,
                'moving_experiment': 345,
                'transform': {'best_registration': ['anything else']}},
            {
                'fixed_experiment': 123,
                'moving_experiment': 345,
                'transform': {'best_registration': ['anything else']}}]
    paths = []
    for i, p in enumerate(pair_outputs):
        paths.append(tmpdir / f"{i}.json")
        with open(paths[-1], "w") as fp:
            json.dump(p, fp)
    df, id_map = utils.summarize_registration_success(paths)
    np.testing.assert_array_equal(df.values, np.array([[1, 0, 1],
                                                       [0, 1, 1],
                                                       [1, 1, 1]]))
    assert id_map == {0: 123, 1: 234, 2: 345}


def test_mask_from_json(experiments):
    for exp in experiments:
        mask, mdict = utils.labeled_mask_from_experiment(exp)
        assert len(mask.shape) == 3
        assert (
                set(np.unique(mask[mask != 0])) ==
                set(np.array(list(mdict.keys())).astype(int)))


def test_create_nice_mask(experiments, tmpdir):
    exp = experiments[0]
    output_dir = str(tmpdir.mkdir("nice_mask"))
    nice_exp = utils.create_nice_mask(exp, output_dir)
    for k in exp:
        assert k in nice_exp
        assert exp[k] == nice_exp[k]
    for k in ['nice_mask_path', 'nice_dict_path']:
        assert k in nice_exp
        assert os .path.isfile(nice_exp[k])


def test_row_col():
    h = 120
    w = 100
    roi = {
            "mask_matrix": np.random.randint(
                0, 2, size=(h, w)).astype(bool),
            "x": np.random.randint(1, 1000),
            "y": np.random.randint(1, 1000),
            "height": h,
            "width": w
            }
    inds = []
    for ix in range(w):
        for iy in range(h):
            if roi['mask_matrix'][iy, ix]:
                inds.append([
                    iy + roi["y"],
                    ix + roi["x"]])
    inds = np.array(inds)
    masked = utils.row_col_from_roi(roi)
    isets = set([tuple(i) for i in inds])
    msets = set([tuple(i) for i in masked])
    assert msets == isets


def test_first_order_properties():
    M = np.eye(3, 3)
    M[0:2, 2] = np.random.randn(2)
    prop = utils.calc_first_order_properties(M)
    assert prop['scale'] == (1.0, 1.0)
    assert np.isclose(prop['shear'], 0.0)
    assert np.all(prop['translation'] == M[0: 2, 2])
    assert np.isclose(prop['rotation'], 0.0)
    assert 'warning' not in prop

    M[2, 1] = 0.1
    prop = utils.calc_first_order_properties(M)
    assert prop['scale'] == (1.0, 1.0)
    assert np.isclose(prop['shear'], 0.0)
    assert np.all(prop['translation'] == M[0: 2, 2])
    assert np.isclose(prop['rotation'], 0.0)
    assert 'warning' in prop


@pytest.mark.parametrize('depth', [1, 4])
def test_read_tiff_3d(depth, tmpdir):
    mask = np.random.randint(0, 500, size=(depth, 100, 100)).astype('uint16')
    output_dir = str(tmpdir.mkdir('tiffs'))
    fname = os.path.join(output_dir, 'tiff_%d.tif' % depth)

    mask_list = [PIL.Image.fromarray(ix) for ix in mask]
    mask_list[0].save(fname, save_all=True, append_images=mask_list[1:])

    new_mask = utils.read_tiff_3d(fname)

    assert new_mask.shape == mask.shape
    assert np.all(new_mask == mask)


@pytest.mark.parametrize(
        "rois, shape, expected",
        [
            (  # 2 overlapping ROIS
                [
                    {
                        'id': 1,
                        'x': 2,
                        'y': 2,
                        'width': 2,
                        'height': 2,
                        'mask_matrix': [[True, True],
                                        [True, True]]
                        },
                    {
                        'id': 2,
                        'x': 3,
                        'y': 3,
                        'width': 2,
                        'height': 2,
                        'mask_matrix': [[True, True],
                                        [True, True]]
                        }],
                (5, 5),
                np.array([
                    [[0, 0, 0, 0, 0],
                     [0, 0, 0, 0, 0],
                     [0, 0, 1, 1, 0],
                     [0, 0, 1, 1, 0],
                     [0, 0, 0, 0, 0]],
                    [[0, 0, 0, 0, 0],
                     [0, 0, 0, 0, 0],
                     [0, 0, 0, 0, 0],
                     [0, 0, 0, 2, 2],
                     [0, 0, 0, 2, 2]]]).astype('uint32')),
            (  # 2 non-overlapping ROIs
                [
                    {
                        'id': 1,
                        'x': 1,
                        'y': 1,
                        'width': 2,
                        'height': 2,
                        'mask_matrix': [[True, True],
                                        [True, True]]
                        },
                    {
                        'id': 2,
                        'x': 3,
                        'y': 3,
                        'width': 2,
                        'height': 2,
                        'mask_matrix': [[True, True],
                                        [True, True]]
                        }],
                (5, 5),
                np.array([
                    [[0, 0, 0, 0, 0],
                     [0, 1, 1, 0, 0],
                     [0, 1, 1, 0, 0],
                     [0, 0, 0, 2, 2],
                     [0, 0, 0, 2, 2]]]).astype('uint32')),
            (  # 3 ROIs destined for 2 layers, out of order
                [
                    {
                        'id': 1,
                        'x': 1,
                        'y': 1,
                        'width': 2,
                        'height': 2,
                        'mask_matrix': [[True, True],
                                        [True, True]]
                        },
                    {
                        'id': 2,
                        'x': 2,
                        'y': 2,
                        'width': 2,
                        'height': 2,
                        'mask_matrix': [[True, True],
                                        [True, True]]
                        },
                    {
                        'id': 3,
                        'x': 3,
                        'y': 3,
                        'width': 2,
                        'height': 2,
                        'mask_matrix': [[True, True],
                                        [True, True]]
                        }],
                (5, 5),
                np.array([
                    [[0, 0, 0, 0, 0],
                     [0, 1, 1, 0, 0],
                     [0, 1, 1, 0, 0],
                     [0, 0, 0, 3, 3],
                     [0, 0, 0, 3, 3]],
                    [[0, 0, 0, 0, 0],
                     [0, 0, 0, 0, 0],
                     [0, 0, 2, 2, 0],
                     [0, 0, 2, 2, 0],
                     [0, 0, 0, 0, 0]]]).astype('uint32')),
            (  # no ROIs
                [],
                (5, 5),
                np.array([
                    [[0, 0, 0, 0, 0],
                     [0, 0, 0, 0, 0],
                     [0, 0, 0, 0, 0],
                     [0, 0, 0, 0, 0],
                     [0, 0, 0, 0, 0]]]).astype('uint32')),
                    ])
def test_layered_mask_from_rois(rois, shape, expected):
    masks = utils.layered_mask_from_rois(rois, shape)
    np.testing.assert_array_equal(masks, expected)



import pytest
import numpy as np
import nway.meta_registration as mr
import nway.transforms as tf


@pytest.mark.parametrize("include_original", [True, False])
def test_MetaRegistration_init(include_original):
    m = mr.MetaRegistration(include_original=include_original)
    names = []
    for candidate in m.candidates:
        assert isinstance(candidate, tf.TransformList)
        names.append([i.__class__.__name__
                      for i in candidate.transforms])
    if include_original:
        assert ['CLAHE', 'ECC'] in names
    if not include_original:
        assert ['CLAHE', 'ECC'] not in names

    assert isinstance(m.contrast, tf.CLAHE)
    assert callable(m.measure)


def test_MetaRegistration_estimate():
    candidates = [
            tf.TransformList(
                transforms=[tf.Crop(edge_buffer=1), tf.Transform()]),
            tf.TransformList(
                transforms=[tf.Crop(edge_buffer=257), tf.Transform()]),
            tf.TransformList(transforms=[tf.Transform(), tf.Transform()])]
    src = np.ones((512, 512)).astype('uint8')
    dst = np.ones((512, 512)).astype('uint8')

    # transforms do not have matrix attribute before estmation
    for candidate in candidates:
        for tform in candidate.transforms:
            assert not hasattr(tform, 'matrix')
    candidates, failed = mr.MetaRegistration.estimate(candidates, src, dst)

    # intentionally made one raise an exception (the large crop)
    assert True in failed

    # successful estimations now have matrix attribute
    for fail, candidate in zip(failed, candidates):
        if not fail:
            for tform in candidate.transforms:
                assert hasattr(tform, 'matrix')


def test_MetaRegistration_evaluate():
    candidates = [
            tf.TransformList(
                transforms=[tf.Crop(edge_buffer=1), tf.Transform()]),
            tf.TransformList(
                transforms=[tf.Crop(edge_buffer=257), tf.Transform()]),
            tf.TransformList(transforms=[tf.Transform(), tf.Transform()])]
    src = np.ones((512, 512)).astype('uint8')
    dst = np.ones((512, 512)).astype('uint8')
    candidates, failed = mr.MetaRegistration.estimate(candidates, src, dst)

    def measure(src, dst):
        return 1.0
    contrast = tf.CLAHE(CLAHE_grid=24, CLAHE_clip=2.5)

    scores = mr.MetaRegistration.evaluate(candidates, failed,
                                          'MOTION_EUCLIDEAN',
                                          contrast, measure, src, dst)
    assert len(scores) == len(candidates)
    for fail, score in zip(failed, scores):
        if fail:
            assert score is None
        else:
            assert score == 1.0


@pytest.mark.parametrize(
        "scores, expected",
        [
            ([0.1, 0.2, 0.3, 0.14], 2),
            ([0.1, 0.2, None, 0.14], 1)])
def test_MetaRegistration_select(scores, expected):
    index = mr.MetaRegistration.select(scores)
    assert index == expected


def test_MetaRegistration_call():
    src = np.ones((512, 512)).astype('uint8')
    dst = np.ones((512, 512)).astype('uint8')

    m = mr.MetaRegistration()
    m(src, dst)
    for att in ['failed', 'scores', 'best_matrix', 'best_candidate']:
        assert hasattr(m, att)



import itertools
import json
import os
import time
from pathlib import Path
from typing import Tuple

import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import PIL.Image
from argschema import ArgSchemaParser
from matplotlib.backends.backend_pdf import PdfPages
from matplotlib.figure import Figure
from skimage.metrics import structural_similarity

import nway.image_processing_utils as imutils
from nway.schemas import NwayDiagnosticSchema, NwayMatchSummarySchema


def cell_experiment_dict(nway_output_path):
    """lookup dict for experiment by cell ID

    Parameters
    ----------
    nway_output_path : str
        path to output json from NwayMatching

    Returns
    -------
    lookup : dict
        dictionary with "cellID": "experimentID"

    """
    with open(nway_output_path, 'r') as f:
        j = json.load(f)
    lookup = {}
    for pw in j['pairwise_results']:
        print(list(pw.keys()))
        for k in ['rejected', 'matches']:
            for pair in pw[k]:
                lookup[pair['fixed']] = pw['fixed_experiment']
                lookup[pair['moving']] = pw['moving_experiment']
        for cellid in pw['unmatched']['fixed']:
            lookup[cellid] = pw['fixed_experiment']
        for cellid in pw['unmatched']['moving']:
            lookup[cellid] = pw['moving_experiment']
    return lookup


def pairwise_transforms(
        nway_output_path, fig=None, subplot_spec=None, fontsize=6):
    """summarize transform parameters and optionally plot

    Parameters
    ----------
    nway_output_path : str
        path to output json from NwayMatching
    fig : :class:`matplotlib.figure.Figure`
        destination figure for plots. No plotting if None
    subplot_spec : :class:`matplotlib.gridspec.SubplotSpec`
        destination SubplotSpec for plots. If None, subplots
        consumer entire figure
    fontsize : int
        fontsize for text in plot

    Returns
    -------
    results : dict
        summarized pairwise transform parameters and experiment ids

    """
    # read the results into lists
    with open(nway_output_path, 'r') as f:
        j = json.load(f)
    pairs = j['pairwise_results']
    props = list(pairs[0]['transform']['properties'].keys())
    results = {}
    for k in props:
        results[k] = [pair['transform']['properties'][k] for pair in pairs]
    results['ids'] = [
            "%d-\n%d" % (
                pair['moving_experiment'],
                pair['fixed_experiment']) for pair in pairs]

    if fig is not None:
        # plot, if a figure is provided
        if subplot_spec is None:
            # these subplots fill entire figure
            outer_plot_grid = gridspec.GridSpec(1, 1, wspace=0.2, hspace=0.2)
            subplot_spec = outer_plot_grid[0]

        inner_plot_grid = gridspec.GridSpecFromSubplotSpec(
                len(props), 1, subplot_spec=subplot_spec,
                wspace=0.1, hspace=0.1)

        x = np.arange(len(results['ids']))
        labels = ['x', 'y']
        for ip, prop in enumerate(props):
            ax = fig.add_subplot(inner_plot_grid[ip])
            ares = np.array(results[prop])
            if len(ares.shape) == 1:
                ares = ares.reshape(-1, 1)
            ares = ares.transpose()
            for iv, v in enumerate(ares):
                ax.plot(x, v, '-o', label=labels[iv])
                ax.set_ylabel(prop, fontsize=fontsize)
            if ares.shape[0] == 2:
                ax.legend(ncol=2, fontsize=fontsize, loc='best', frameon=False)
            ax.set_xticks(x)
            ax.set_xticklabels([])
            if ip == 0:
                ax.set_title(
                        'pairwise transform properties', fontsize=fontsize)
        ax.set_xticklabels(results['ids'], rotation=45, fontsize=fontsize)
        ax.set_xlabel('pairwise experiment IDs', fontsize=fontsize)

    return results


def some_grid(n):
    """specify a roughly square grid with n elements

    Parameters
    ----------
    n : int
        number of elements

    Returns
    -------
    (nrow, ncol) : tuple
        shape for new grid

    """
    nrow = int(np.floor(np.sqrt(n)))
    # NOTE float cast for python 2.7
    ncol = int(np.ceil(float(n) / nrow))
    return (nrow, ncol)


def pairwise_matches(
        nway_output_path, fig=None, subplot_spec=None, fontsize=6):
    """summarize pairwise matches and reject costs

    Parameters
    ----------
    nway_output_path : str
        path to output json from NwayMatching
    fig : :class:`matplotlib.figure.Figure`
        destination figure for plots. No plotting if None
    subplot_spec : :class:`matplotlib.gridspec.SubplotSpec`
        destination SubplotSpec for plots. If None, subplots
        consumer entire figure
    fontsize : int
        fontsize for text in plot

    Returns
    -------
    costs : dict
        costs[pair id] = {
            'matches': list of costs,
            'rejected': list of costs}
    allcosts = dict
        'matches' : concat of all pair matches,
        'rejected' : concate of all pair rejected

    """
    # read the results into lists
    with open(nway_output_path, 'r') as f:
        j = json.load(f)
    pairs = j['pairwise_results']

    costs = {}
    subkeys = ['matches', 'rejected']
    for pair in pairs:
        k = "%d-%d" % (pair['moving_experiment'], pair['fixed_experiment'])
        costs[k] = {}
        for subk in subkeys:
            costs[k][subk] = [m['cost'] for m in pair[subk]]
    allcosts = {}
    for subk in subkeys:
        allcosts[subk] = np.concatenate([c[subk] for c in costs.values()])

    if fig is not None:
        # plot, if a figure is provided
        if subplot_spec is None:
            # these subplots fill entire figure
            outer_plot_grid = gridspec.GridSpec(1, 1, wspace=0.2, hspace=0.2)
            subplot_spec = outer_plot_grid[0]

        inner_plot_grid = gridspec.GridSpecFromSubplotSpec(
                *some_grid(len(costs)), subplot_spec=subplot_spec,
                wspace=0.1, hspace=0.1)

        bins = np.arange(0, 2.1, 0.1)

        for ic, ck in enumerate(costs):
            ax = fig.add_subplot(inner_plot_grid[ic])
            ax.hist(
                    allcosts['matches'],
                    histtype='step',
                    color='k',
                    bins=bins,
                    label='all matches')
            ax.hist(
                    allcosts['rejected'],
                    histtype='step',
                    color='r',
                    bins=bins,
                    label='all rejected')
            ax.hist(
                    costs[ck]['matches'],
                    color='k',
                    alpha=0.5,
                    bins=bins,
                    label='pair matches')
            ax.hist(
                    costs[ck]['rejected'],
                    color='r',
                    alpha=0.5,
                    bins=bins,
                    label='pair rejected')
            ax.legend(loc=2, fontsize=fontsize, title=ck, frameon=False)
            ax.get_legend().get_title().set_fontsize(fontsize)
            ax.set_yscale('log')
            ax.set_xlabel('cost', fontsize=6)
            ax.set_ylabel('match count', fontsize=6)
            rc = inner_plot_grid[ic].get_rows_columns()
            if rc[3] != (rc[0] - 1):
                ax.set_xticks([])
            if rc[5] != 0:
                ax.set_yticks([])

    return costs, allcosts


def nway_matches(
        nway_output_path, fig=None, fontsize=12):
    """summarize pairwise matches and reject costs

    Parameters
    ----------
    nway_output_path : str
        path to output json from NwayMatching
    fig : :class:`matplotlib.figure.Figure`
        destination figure for plots. No plotting if None
    fontsize : int
        fontsize for text in plot

    Returns
    -------
    allnw : dict
        keys are sorted tuples of match sets
        values are :
            n : number of pairwise matches (< max_distance)
            avecost : average pairwise cost

    """

    _, allcosts = pairwise_matches(nway_output_path)

    # read the results into lists
    with open(nway_output_path, 'r') as f:
        j = json.load(f)
    pairs = j['pairwise_results']
    nway = j['nway_matches']

    # make a set of tuples of all pairwise matches and rejected
    # (anything within max_distance)
    # keep the tuples ordered so we can search easily
    allpw = {}
    for pair in pairs:
        for subk in ['matches', 'rejected']:
            pw = {tuple(np.sort([m['moving'], m['fixed']])): m['cost']
                  for m in pair[subk]}
            allpw.update(pw)

    # go through all the nway matches and get the n and mean for each
    # entry
    allnw = {}
    for match in nway:
        k = tuple(match)
        n = 0
        avcost = 0
        for candidate in itertools.combinations(match, 2):
            tsort = tuple(np.sort(candidate))
            if tsort in allpw:
                n += 1
                avcost += allpw[tsort]
        if n != 0:
            allnw[k] = {
                    'n': n,
                    'avecost': avcost / n}

    if fig is not None:
        spec = gridspec.GridSpec(ncols=5, nrows=5, figure=fig)

        ax0 = fig.add_subplot(spec[1:, 0:4])
        ax_histy = fig.add_subplot(spec[1:5, 4:])
        ax_histx = fig.add_subplot(spec[0, 0:4])

        ns = np.array([i['n'] for i in allnw.values()])
        x = np.arange(ns.min(), ns.max() + 1)
        bins = np.arange(0.0, 2.1, 0.1)
        y = []
        cbinned = []
        for ix in x:
            iy = [i['avecost'] for i in allnw.values() if i['n'] == ix]
            y.append(iy)
            h, _ = np.histogram(iy, bins=bins)
            cbinned.append(h)
        cbinned = np.array(cbinned)

        ax0.imshow(
                np.flipud(cbinned),
                extent=[
                    bins.min(),
                    bins.max(),
                    ns.min() - 0.5,
                    ns.max() + 0.5],
                cmap='gray_r',
                aspect='auto')

        xbins = [bins[i:i+2].mean() for i in range(bins.size - 1)]
        ax_histx.bar(
                xbins,
                cbinned.sum(axis=0),
                color='k',
                alpha=0.5,
                width=0.1,
                edgecolor='k')
        ax_histy.barh(
                x,
                cbinned.sum(axis=1),
                color='k',
                alpha=0.5,
                height=1.0,
                edgecolor='k')

        ax0.set_xlabel('cost', fontsize=fontsize)
        ax0.set_ylabel('matches per set', fontsize=fontsize)
        ax_histx.set_xticks([])
        ax_histy.set_yticks([])
        ax_histy.set_xlabel('count', fontsize=fontsize)
        ax_histx.set_ylabel('count', fontsize=fontsize)
        ax_histx.set_title('nway match results', fontsize=fontsize)
        ax_histx.set_xlim(ax0.get_xlim())
        ax_histy.set_ylim(ax0.get_ylim())

    return allnw


def plot_all(nway_output_path, fname=None):
    fs = (12, 8)

    fig1 = plt.figure(clear=True, figsize=fs)
    pairwise_transforms(nway_output_path, fig=fig1)

    fig2 = plt.figure(clear=True, figsize=fs)
    pairwise_matches(nway_output_path, fig=fig2)

    fig3 = plt.figure(clear=True, figsize=fs)
    nway_matches(nway_output_path, fig=fig3)

    if fname is not None:
        p = PdfPages(fname)
        p.savefig(fig1)
        p.savefig(fig2)
        p.savefig(fig3)
        p.close()


def create_nway_input_maps(nway_input: dict) -> Tuple[dict, dict]:
    """Create mappings between experiment id, experiment stimulus name,
    and the average intensity projection image arrays.

    Parameters
    ----------
    nway_input : dict
        The 'input.json' in dictionary form passed to the nway matching
        module.

    Returns
    -------
    Tuple[dict, dict]
        A tuple of two mappings. The first mapping relates experiment ids to
        experiment stimulus names. The second mapping relates experiment ids
        to the average intensity projection image array for the experiment.
    """
    ophys_expts = nway_input['experiment_containers']['ophys_experiments']

    expt_id_stim_name_map = dict()
    expt_id_avg_image_map = dict()
    for expt in ophys_expts:
        stimulus_name = expt.get('stimulus_name')
        if stimulus_name is None:
            stimulus_name = 'Unknown Stimulus'
        expt_id_stim_name_map[expt['id']] = stimulus_name

        avg_image_key = 'ophys_average_intensity_projection_image'
        with PIL.Image.open(expt[avg_image_key]) as im:
            expt_avg_image = np.array(im)
        expt_id_avg_image_map[expt['id']] = expt_avg_image

    expt_id_stim_name_map = {k: v for k, v
                             in sorted(expt_id_stim_name_map.items(),
                                       key=lambda x: str(x[1]))}

    return (expt_id_stim_name_map, expt_id_avg_image_map)


def create_nway_summary_df(expt_id_stim_name_map: dict,
                           expt_id_avg_image_map: dict,
                           nway_output: dict) -> pd.DataFrame:
    """Create an nway matching summary dataframe necessary for plotting
    match fractions and assessing average image registrations.

    Parameters
    ----------
    expt_id_stim_name_map : dict
        A mapping that relates experiment ids to experiment stimulus names.
        Produced by 'create_nway_input_maps'.
    expt_id_avg_img_map : dict
        A mapping that relates experiment ids to the average intensity
        projection image array for the experiment. Produced by
        'create_nway_input_maps'.
    nway_output : dict
        The 'output.json' in dictionary form produced by the nway matching
        module.

    Returns
    -------
    pd.DataFrame
        A pandas DataFrame with the following columns:

        fixed_expt (int): The experiment id of the alignment target
        fixed_expt_stim_name (str): The stimulus name for the fixed expt
        moving_expt (int): The experiment id of the image to align
        moving_expt_stim_name (str): The stimulus name for the moving expt
        n_unmatched_fixed (int): Number of ROIs from the fixed experiment which
            could not be matched to a moving experiment's ROI
        n_unmatched_moving (int): Number of ROIs from the moving experiment
            which could not be matched to a fixed experiment's ROI
        n_matches (int): Number of ROIs that were matched between fixed and
            moving experiments
        n_total (int): The total number of ROIs to match
        fraction_matched (float): n_matches divided by n_total

        This DataFrame additionally contains the following attributes keys:

        warped_images: A mapping that relates a pairwise match to a warped
            registration image
        expt_id_stim_name_map: A mapping that relates an experiment's id with a
            stimulus name (describing the type of experiment)
        expt_id_avg_image_map: A mapping that relates an experiment's id with
            the experiment's average intensity projection image array
    """
    pairwise_results = nway_output['pairwise_results']

    warped_avg_image_maps = dict()
    df_list = []
    for pair in pairwise_results:
        # Assemble match statistics
        fixed_expt = pair['fixed_experiment']
        fixed_expt_stim_name = expt_id_stim_name_map[fixed_expt]

        moving_expt = pair['moving_experiment']
        moving_expt_stim_name = expt_id_stim_name_map[moving_expt]

        n_unmatched_fixed = len(pair['unmatched']['fixed'])
        n_unmatched_moving = len(pair['unmatched']['moving'])
        n_matches = len(pair['matches'])
        n_total = n_matches + n_unmatched_fixed + n_unmatched_moving
        fraction_matched = n_matches / float(n_total)
        df_list.append([fixed_expt, fixed_expt_stim_name, moving_expt,
                        moving_expt_stim_name, n_unmatched_fixed,
                        n_unmatched_moving, n_matches, n_total,
                        fraction_matched])

        # Recreate warped images
        transform_matrix = np.array(pair['transform']['matrix'])
        transform_type = pair['transform']['transform_type']
        fixed_img = expt_id_avg_image_map[fixed_expt]
        moving_img = expt_id_avg_image_map[moving_expt]

        moving_warped = imutils.warp_image(moving_img, transform_matrix,
                                           transform_type, fixed_img.shape)
        warped_img_key_1 = f"{moving_expt}_to_{fixed_expt}"
        warped_avg_image_maps[warped_img_key_1] = moving_warped

        inv_transform = np.linalg.inv(transform_matrix)
        fixed_warped = imutils.warp_image(fixed_img, inv_transform,
                                          transform_type, moving_img.shape)
        warped_img_key_2 = f"{fixed_expt}_to_{moving_expt}"
        warped_avg_image_maps[warped_img_key_2] = fixed_warped

    columns = ['fixed_expt', 'fixed_expt_stim_name', 'moving_expt',
               'moving_expt_stim_name', 'n_unmatched_fixed',
               'n_unmatched_moving', 'n_matches', 'n_total',
               'fraction_matched']
    nway_summary_df = pd.DataFrame(df_list, columns=columns)
    nway_summary_df = nway_summary_df.sort_values(by=['fixed_expt_stim_name',
                                                      'moving_expt_stim_name'],
                                                  ignore_index=True)
    nway_summary_df.attrs['warped_images'] = warped_avg_image_maps
    nway_summary_df.attrs['expt_id_stim_name_map'] = expt_id_stim_name_map
    nway_summary_df.attrs['expt_id_avg_image_map'] = expt_id_avg_image_map

    return nway_summary_df


def plot_container_match_fraction(nway_summary_df: pd.DataFrame) -> Figure:
    """Given an nway summary DataFrame, produce a plot summarizing ROI
    match fractions.
    """
    expt_id_stim_name_map = nway_summary_df.attrs['expt_id_stim_name_map']
    expt_ids = expt_id_stim_name_map.keys()

    match_frac_mtx = pd.DataFrame(index=expt_ids,
                                  columns=expt_ids,
                                  dtype=float)
    for _, row in nway_summary_df.iterrows():
        match_fraction = row['fraction_matched']
        match_frac_mtx[row['fixed_expt']][row['moving_expt']] = match_fraction
        match_frac_mtx[row['moving_expt']][row['fixed_expt']] = match_fraction
    np.fill_diagonal(match_frac_mtx.values, 1.0)

    fig, ax = plt.subplots(figsize=(12, 12))
    mat_ax = ax.matshow(match_frac_mtx, vmin=0.0, vmax=1.0, cmap='magma')

    for (i, j), data in np.ndenumerate(match_frac_mtx.values):
        if i == j:
            ax.text(j, i, f'{data:.3f}', ha='center', va='center',
                    fontsize=18)
        else:
            ax.text(j, i, f'{data:.3f}', ha='center', va='center',
                    fontsize=18, color='white')

    ax.xaxis.set_ticks_position('bottom')
    xy_labels = [f"{stim_name} (Expt: {expt_id})"
                 for expt_id, stim_name in expt_id_stim_name_map.items()]
    ax.set_xticks(list(range(len(expt_ids))))
    ax.set_xticklabels(xy_labels, fontsize=18, rotation=45, ha='right')
    ax.set_yticks(list(range(len(expt_ids))))
    ax.set_yticklabels(xy_labels, fontsize=18)

    plt.title("Fraction matched ROIs across sessions", fontsize=24, pad=20)
    fig.colorbar(mat_ax)
    return fig


def plot_container_warp_overlays(nway_summary_df: pd.DataFrame) -> Figure:
    """Given an nway summary DataFrame, produce a plot that shows the
    overlap of experiment average intensity projection images after
    registration.
    """
    expt_id_stim_name_map = nway_summary_df.attrs['expt_id_stim_name_map']
    expt_ids = expt_id_stim_name_map.keys()

    expt_id_avg_image_map = nway_summary_df.attrs['expt_id_avg_image_map']
    warped_images_map = nway_summary_df.attrs['warped_images']

    panel_len = len(expt_ids) + 1

    fig, axes = plt.subplots(nrows=panel_len,
                             ncols=panel_len,
                             figsize=(25, 25))
    # Turn off all axes in subplots
    for ax in axes.ravel():
        ax.set_axis_off()

    # plot unwarped base images
    for idx, _id in enumerate(expt_ids, start=1):
        session_type = '_'.join(expt_id_stim_name_map[_id].split('_')[:2])

        axes[0][idx].imshow(expt_id_avg_image_map[_id], cmap='gray')
        axes[0][idx].set_title(f"Expt: {_id}\n{session_type}", fontsize=18)
        axes[idx][0].imshow(expt_id_avg_image_map[_id], cmap='gray')
        axes[idx][0].set_title(f"Expt: {_id}\n{session_type}",
                               x=-0.5, y=0.4, fontsize=18)

    # plot warped 'moving' expt image on 'fixed' expt image
    for row, expt_1 in enumerate(expt_ids, start=1):
        for col, expt_2 in enumerate(expt_ids, start=1):
            if expt_1 == expt_2:
                continue
            else:
                expt_1_avg_img = expt_id_avg_image_map[expt_1]
                norm_expt_1_avg_img = (
                    expt_1_avg_img / float(np.amax(expt_1_avg_img)))

                warp_key = f"{expt_2}_to_{expt_1}"
                warped_avg_img = warped_images_map[warp_key]

                norm_warped_avg_img = (
                    warped_avg_img / float(np.amax(warped_avg_img)))

                img_shape = norm_expt_1_avg_img.shape
                combined_img = np.zeros((img_shape[0], img_shape[1], 3))
                combined_img[:, :, 0] = norm_expt_1_avg_img
                combined_img[:, :, 1] = norm_warped_avg_img

                axes[row][col].imshow(combined_img)
                ssim = structural_similarity(norm_expt_1_avg_img,
                                             norm_warped_avg_img,
                                             gaussian_weights=True)
                axes[row][col].set_title(f"SSIM: {ssim:.3f}", fontsize=16)

    fig.tight_layout()
    return fig


def plot_container_warp_summary(nway_summary_df: pd.DataFrame) -> Figure:
    """Given an nway summary DataFrame, produce a plot that shows in
    greater detail the quality of the registration between experiment
    average intensity projection images.
    """
    expt_id_avg_image_map = nway_summary_df.attrs['expt_id_avg_image_map']
    warped_images_map = nway_summary_df.attrs['warped_images']

    num_ax_cols = len(nway_summary_df.index)

    fig, axes = plt.subplots(nrows=4,
                             ncols=num_ax_cols,
                             figsize=(30, 10))
    # Turn off all axes in subplots
    for ax in axes.ravel():
        ax.set_axis_off()

    for idx, row in nway_summary_df.iterrows():
        fixed_expt = row['fixed_expt']
        moving_expt = row['moving_expt']

        moving_image = expt_id_avg_image_map[moving_expt]
        fixed_image = expt_id_avg_image_map[fixed_expt]
        warp_key = f"{moving_expt}_to_{fixed_expt}"
        warped_image = warped_images_map[warp_key]

        norm_fixed_image = fixed_image / float(np.amax(fixed_image))
        norm_warped_image = warped_image / float(np.amax(warped_image))
        img_shape = norm_fixed_image.shape
        combined_img = np.zeros((img_shape[0], img_shape[1], 3))
        combined_img[:, :, 0] = norm_fixed_image
        combined_img[:, :, 1] = norm_warped_image

        ssim = structural_similarity(fixed_image, warped_image,
                                     gaussian_weights=True)

        moving_stimulus_name = row['moving_expt_stim_name']
        moving_session_type = '_'.join(moving_stimulus_name.split('_')[:2])
        fixed_stimulus_name = row['fixed_expt_stim_name']
        fixed_session_type = '_'.join(fixed_stimulus_name.split('_')[:2])

        axes[0][idx].imshow(moving_image, cmap='gray')
        axes[0][idx].set_title(f"Moving\n{moving_expt}\n{moving_session_type}")

        axes[1][idx].imshow(fixed_image, cmap='gray')
        axes[1][idx].set_title(f"Fixed\n{fixed_expt}\n{fixed_session_type}")

        axes[2][idx].imshow(warped_image, cmap='gray')
        axes[2][idx].set_title(f"Registered\n{moving_expt}")

        axes[3][idx].imshow(combined_img)
        axes[3][idx].set_title(f"SSIM\n{ssim:.3f}", fontsize=16)

    fig.tight_layout()
    return fig


class NwaySummary(ArgSchemaParser):
    default_schema = NwayMatchSummarySchema

    def run(self) -> dict:
        input_maps = create_nway_input_maps(self.args['nway_input'])
        expt_id_stim_name_map, expt_id_avg_image_map = input_maps

        summary_df = create_nway_summary_df(expt_id_stim_name_map,
                                            expt_id_avg_image_map,
                                            self.args['nway_output'])

        fig1 = plot_container_match_fraction(summary_df)
        fig2 = plot_container_warp_overlays(summary_df)
        fig3 = plot_container_warp_summary(summary_df)

        save_dir = Path(self.args['output_directory'])
        timestamp = time.strftime('%Y_%m_%d_%H_%M_%S', time.localtime())

        fig1_save_path = save_dir / f"nway_match_fraction_plot_{timestamp}.png"
        fig2_save_path = save_dir / f"nway_warp_overlay_plot_{timestamp}.png"
        fig3_save_path = save_dir / f"nway_warp_summary_plot_{timestamp}.png"

        fig1.savefig(fig1_save_path, dpi=300, bbox_inches="tight")
        fig2.savefig(fig2_save_path, dpi=300, bbox_inches="tight")
        fig3.savefig(fig3_save_path, dpi=300, bbox_inches="tight")

        return {
            "nway_match_fraction_plot": str(fig1_save_path),
            "nway_warp_overlay_plot": str(fig2_save_path),
            "nway_warp_summary_plot": str(fig3_save_path)
        }


class NwayDiagnostics(ArgSchemaParser):
    default_schema = NwayDiagnosticSchema

    def run(self):
        if self.args['use_input_dir']:
            self.args['output_pdf'] = os.path.join(
                    os.path.dirname(self.args['input_json']),
                    os.path.basename(self.args['output_pdf']))

        plot_all(self.args['input_json'], fname=self.args['output_pdf'])


if __name__ == "__main__":  # pragma: no cover
    nd = NwayDiagnostics()
    nd.run()



import nway.image_processing_utils as imu
import pytest
from pathlib import Path
import numpy as np
import PIL.Image

TEST_FILE_DIR = Path.resolve(Path(__file__)).parent / "test_files"


@pytest.fixture
def fixed_image_fixture():
    # grab one specific file because failing large translation
    # cases vary image-by-image
    fgen = TEST_FILE_DIR.rglob("avgInt_a1X.png")
    for f in fgen:
        if "ophys_session_775289198" in str(f):
            impath = f
    with PIL.Image.open(impath) as fp:
        image = np.array(fp)
    return image


@pytest.mark.parametrize("CLAHE_GRID, CLAHE_CLIP",
                         [(8, 2.5), (24, 2.5), (-1, 2.5)])
def test_contrast_adjust(fixed_image_fixture, CLAHE_GRID, CLAHE_CLIP):
    ca_image = imu.contrast_adjust(fixed_image_fixture, CLAHE_GRID, CLAHE_CLIP)
    assert ca_image.shape == fixed_image_fixture.shape
    assert ca_image.dtype == fixed_image_fixture.dtype


@pytest.mark.parametrize("motion_type", [
                         "MOTION_TRANSLATION", "MOTION_EUCLIDEAN",
                         "MOTION_AFFINE", "MOTION_HOMOGRAPHY"])
def test_warp_image(fixed_image_fixture, motion_type):
    tform = np.array([[1.1, 0.1, 12.3],
                      [-0.5, 0.94, -23],
                      [0.0, 0.0, 1.0]])
    warped = imu.warp_image(
            fixed_image_fixture, tform, motion_type, fixed_image_fixture.shape)
    assert warped.shape == fixed_image_fixture.shape
    assert warped.dtype == 'uint8'



import numpy as np
from skimage.metrics import structural_similarity
from functools import partial
from typing import List, Union, Callable, Tuple
import nway.image_processing_utils as impu
import nway.transforms as tf


class MetaRegistration():
    """Attempts multiple types of registration and chooses the best
    one based on an evaluation metric.

    Attributes
    ----------
    candidates: List[tf.TransformList]
        the candidate registration pipelines. Each one is a TransformList
    scores: List[float]
        the evaluation scores associated with each candidate
    transformed_images: List[List[np.ndarray]]
        each entry is the resulting transformed images from the candidate
        pipelines. NOTE: both src and dst can be modified by a transform.
    best_score: float
        the best score of all candidates
    best_matrix: np.ndarray
        the accumulated transform matrix for the best pipeline
    best_candidate: List[str]
        list of names of the transforms from the best candidate

    Methods
    -------
    estimate(src, dst):
        estimates the transforms for all the candidates
    evaluates(src, dst):
        applies the transforms for all the candidates and scores
    select():
        evaluates the scoring of the candidates and sets best attributes
    measure(src, dst):
        perform metric evaluation on a pair of images

    Notes
    -----
    This meta-registration approach accounts for a diverse set of input
    images that were difficult to align with a single strategy. If a new set
    of images is not successfully registered by one of these candidates,
    additional candidate lists can be added, as well as additional Transforms

    """
    def __init__(self, maxCount: int = 1000, epsilon: float = 1.5e-7,
                 motion_type: str = 'MOTION_EUCLIDEAN',
                 gaussFiltSize: int = 5, CLAHE_grid: int = 24,
                 CLAHE_clip: float = 2.5, edge_buffer: int = 40,
                 ssim_sigma: float = 20.0, include_original: bool = False):
        """

        Parameters
        ----------
        maxCount: int
            passed as termination criterion for cv2.findTransformECC
        epsilon: float
            passed as termination criterion for cv2.findTransformECC
        motion_type: str
            used to specify motion_type (only cv2.findTransformECC
            contributes non-Euclidean matrix elements)
        gaussFiltSize: int
            must be odd. passed to cv2.findTransformECC
        CLAHE_grid: int
            passed as tileGridSize to cv2.createCLAHE.
        CLAHE_clip : float
            passed as clipLimit to cv2.createCLAHE
        edge_buffer: int
            this many pixels will be cropped from top, bottom, left, and right
            of both images in the crop() steps.
        ssim_sigma: float
            passed to skimage.metrics.structural_similarity as sigma, with
            gaussian_weights=True. Larger number dampens high frequency
            content of this measure and is thus less sensitive to
            real biological variations.
        include_original: bool
            whether to include the original strategy of CLAHE + ECC which
            occasionally gave a slight registration improvement, but not
            very good. Including for legacy (and legacy regression test)

        """

        # make these partials, so we can succinctly instantiate new ones
        contrast = partial(tf.CLAHE, CLAHE_grid=CLAHE_grid,
                           CLAHE_clip=CLAHE_clip)
        phase = partial(tf.PhaseCorrelate)
        crop = partial(tf.Crop, edge_buffer=edge_buffer)
        ecc = partial(tf.ECC, maxCount=maxCount, epsilon=epsilon,
                      motion_type=motion_type, gaussFiltSize=gaussFiltSize)

        # instantiate one contrast() as attribute for using in evaluate()
        self.contrast = contrast()

        self.motion_type = motion_type

        # function to evaluate alignment
        self.measure = partial(structural_similarity,
                               gaussian_weights=True,
                               sigma=ssim_sigma)

        # in the event that no transform is the best alignment (unlikely)
        # this just makes the reported name `Identity` rather than `Transform`
        class Identity(tf.Transform):
            pass

        # independently attempt alignment multiple ways
        self.candidates = [
                # Identity is unlikely as the best candidate
                # and probably indicates some data problem.
                tf.TransformList(transforms=[Identity()]),
                # cropping eliminates some motion-correction induced
                # borders that can bias the registration
                tf.TransformList(transforms=[crop(), ecc()]),
                # sometimes, ECC works better after contrast adjustment
                tf.TransformList(transforms=[crop(), contrast(), ecc()]),
                # sometimes, just PhaseCorrelation is a good estimator
                tf.TransformList(transforms=[crop(), contrast(), phase()]),
                # sometimes, using the PhaseCorrelation translations
                # as a precursor to ECC is helpful
                tf.TransformList(transforms=[crop(), contrast(), phase(),
                                             ecc()])
                ]
        if include_original:
            self.candidates.append(
                    tf.TransformList(transforms=[contrast(), ecc()]))

    def __call__(self, src, dst):
        """attempts all registration candidates, evaluates them,
        and selects the best based on measure()

        Parameters
        ----------
        src: np.ndarray
            uint8, needed for opencv operations
        dst: np.ndarray
            uint8, needed for opencv operations

        """
        self.candidates, self.failed = self.estimate(self.candidates, src, dst)
        self.scores = self.evaluate(self.candidates, self.failed,
                                    self.motion_type, self.contrast,
                                    self.measure, src, dst)
        index = self.select(self.scores)
        self.best_matrix = self.candidates[index].matrix
        self.best_candidate = [tf.__class__.__name__
                               for tf in self.candidates[index].transforms]

    @staticmethod
    def estimate(candidates, src, dst
                 ) -> Tuple[List[tf.TransformList], List[bool]]:
        """estimates the transforms in each candidate TransformList
        in the event of an exception, notes that candidate as failed.

        Parameters
        ----------
        candidates: List[TransformList]
            candidate TransformLists to evaluate
        src: np.ndarray
            uint8, needed for opencv operations
        dst: np.ndarray
            uint8, needed for opencv operations

        Returns
        -------
        candidates: List[TransformList]
            candidate TransformLists, which have been evaluated
        failed: List[bool]
            same length/order as candidates. Whether the evaluation failed

        """
        failed = [False] * len(candidates)
        for ic, candidate in enumerate(candidates):
            try:
                candidate.estimate(src, dst)
            except tf.TransformException:
                failed[ic] = True
        return candidates, failed

    @staticmethod
    def evaluate(candidates: List[tf.TransformList], failed: List[bool],
                 motion_type: str, contrast: tf.CLAHE,
                 measure: Callable[[np.ndarray, np.ndarray], float],
                 src: np.ndarray, dst: np.ndarray) -> List[Union[float, None]]:
        """evaluates the candidate TransformLists by applying their
        estimated transforms, warping the src image and measuring the
        ssim parameter.

        Parameters
        ----------
        candidates: List[TransformList]
            candidate TransformLists to evaluate
        failed: List[bool]
            whether a particular candidate failed to estimate
        motion_type: str
            motion_type for warping moving image
        contrast: tf.CLAHE
            an instance of CLAHE to apply before warping and evaluating
        measure: function
            evaluates the metric of two images
        src: np.ndarray
            uint8, needed for opencv operations
        dst: np.ndarray
            uint8, needed for opencv operations

        Returns
        -------
        scores: List[float]
             the evaluation scores. Can be None for a candidate that
             raised an exception.

        Note
        ----
        evaluations can only be compared across candidates
        by comparing ssim after geometrical transforming only.
        Attempting to compare ssim between the entire transforms
        will likely not work well, as things like contrast adjustment
        and cropping will change the measure differentially.

        """
        scores = []
        for candidate, ifail in zip(candidates, failed):
            if ifail:
                scores.append(None)
            else:
                mov, fix = contrast.apply(src, dst)
                mov = impu.warp_image(mov, candidate.matrix,
                                      motion_type, fix.shape)
                scores.append(measure(mov, fix))
        return scores

    @staticmethod
    def select(scores: List[Union[float, None]]) -> int:
        """select the best candidate based on the maximum
        ssim metric.

        Parameters
        ----------
        scores: List[float]
            scores from evaluating the candidates. Can contain None

        Returns
        -------
        index: int
            index of best score

        """
        index = np.nanargmax(np.array(scores, dtype='float'))
        return index



import pytest
import nway.nway_matching as nway
import os
from jinja2 import Template
import json
import numpy as np
import copy


TEST_FILE_DIR = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        'test_files')


@pytest.fixture(scope='module')
def table_to_prune():
    fname = os.path.join(
            TEST_FILE_DIR,
            "test0",
            "table_before_pruning.json")
    with open(fname, 'r') as f:
        j = json.load(f)
    table = np.array(j['table'])
    score = np.array(j['score'])
    yield (table, score)


@pytest.fixture(scope='function')
def input_file(tmpdir):
    thistest = os.path.join(TEST_FILE_DIR, 'test0')
    myinput = os.path.join(thistest, 'input.json')
    with open(myinput, 'r') as f:
        template = Template(json.dumps(json.load(f)))
    output_dir = str(tmpdir.mkdir("nway_test"))
    rendered = json.loads(
            template.render(
                output_dir=output_dir,
                test_files_dir=str(thistest)))
    rendered['log_level'] = "DEBUG"
    yield rendered


def test_nway_exception(tmpdir, input_file):
    args = copy.deepcopy(input_file)
    args['experiment_containers']['ophys_experiments'] = \
        [args['experiment_containers']['ophys_experiments'][0]]
    output_dir = str(tmpdir.mkdir("nway_exception"))
    args['output_directory'] = output_dir
    args['assignment_solver'] = 'Blossom'

    nwmatch = nway.NwayMatching(input_data=args, args=[])
    with pytest.raises(nway.NwayException):
        nwmatch.run()


@pytest.mark.parametrize('greduce', ['keepmin', 'popmax'])
def test_default_nway(tmpdir, input_file, greduce):
    args = copy.deepcopy(input_file)
    output_dir = str(tmpdir.mkdir("nway_default"))
    args['output_json'] = os.path.join(output_dir, 'output.json')
    args['pruning_method'] = greduce
    args['include_original'] = True
    n = nway.NwayMatching(input_data=args, args=[])

    assert n.args['assignment_solver'] == 'Blossom'

    n.run()

    with open(n.args['output_json'], 'r') as f:
        oj = json.load(f)

    assert len(oj['nway_matches']) > 650
    nave = np.array([len(i) for i in oj['nway_matches']]).mean()
    assert nave > 1.5

    nexp = len(args['experiment_containers']['ophys_experiments'])

    npairs = int(nexp * (nexp - 1) / 2)

    assert npairs == len(oj['pairwise_results'])



import numpy as np
from skimage.transform import AffineTransform
import PIL.Image
import os
import json
import pandas as pd
import matplotlib.pyplot as plt
from pathlib import Path
from typing import List, Any, Tuple, Optional


def summarize_registration_success(pair_outputs):
    """checks which pairs succeeded in registratio. tries to keep
    dataframe small for logging by separating column/index names
    and returning an idmap also

    Parameters
    ----------
    pair_outputs: List[dict]
        at minimum, each entry is {
                                     'fixed_experiment': <int>,
                                     'moving_experiment': <int>,
                                     'transform': {
                                                    'best_registration': []
                                                  }
                                   }

    Returns
    -------
    df: pd.DataFrame
        boolean-valued dataframe, sorted both in rows and columns
        row/column names replaced with integers
    id_map: dict
        map of row/column names to experiment ids. can be used to reconstruct
        named rows and columns with:
        df = df.rename(index=id_map, columns=id_map)

    """
    df = pd.DataFrame()
    for pair_output in pair_outputs:
        with open(pair_output, "r") as f:
            j = json.load(f)
        success = 1
        if j['transform']['best_registration'] == ['Identity']:
            success = 0
        df.loc[j['fixed_experiment'], j['moving_experiment']] = success
        df.loc[j['moving_experiment'], j['fixed_experiment']] = success
    df = df.sort_index(0)
    df = df.sort_index(1)
    id_map = {v: i for i, v in enumerate(df.index)}
    for v in id_map.keys():
        df.loc[v, v] = 1
    df = df.rename(index=id_map, columns=id_map)
    df = df.astype(int)
    id_map = {v: k for k, v in id_map.items()}
    return df, id_map


def write_pair_images(output_path: Path, titles: List[str],
                      images: List[np.ndarray],
                      suptitle: Optional[str] = None) -> Path:
    """make a plot of images and write to disk

    Parameters
    ----------
    output_path: Path
        destination file
    titles: List[str]
        titles for the subplots
    images: List[np.ndarray]
        images for the subplots
    suptitle: str
        supertitle for the figure

    Return
    ------
    output_path: Path
        output path

    """
    if len(titles) != len(images):
        raise ValueError(f"length of titles {len(titles)} does not match "
                         f"length of images {len(images)}.")
    f, a = plt.subplots(1, len(images), clear=True, num=1, figsize=(12, 8))
    for ax, image, title in zip(a, images, titles):
        ax.imshow(image, cmap='gray')
        ax.set_title(title)
    if suptitle is not None:
        f.suptitle(suptitle)
    f.savefig(output_path)
    return output_path


def create_nice_mask(experiment, output_directory):
    """create a labeled mask from the input json experiment and a translation
    dict to cell IDs, save them to disk, and add their paths to the experiment
    json.

    Parameters
    ----------
    experiment : dict
        output from segmentation
        'ophys_average_intensity_projection_image' : str(path)
        'cell_rois' : dict
        'id' :  str
        'stimulus_name' : str/Null
    output_directory : str
        destination directory for writing masks and translation dicts

    Returns
    -------
    experiment : dict
        same as input with 2 additional key/value pairs
        'nice_mask_path' : str(path)
        'nice_dict_path' : str(path)

    """

    nice_mask, mask_dict = labeled_mask_from_experiment(experiment)

    mask_path = os.path.join(
            output_directory,
            "%d_nice_mask.tif" % experiment['id'])

    nice_mask = [PIL.Image.fromarray(ix) for ix in nice_mask]
    nice_mask[0].save(mask_path, save_all=True, append_images=nice_mask[1:])

    experiment['nice_mask_path'] = mask_path

    dict_path = os.path.join(
            output_directory,
            "%d_nice_dict.json" % experiment['id'])
    full_dict = {
            'experiment': experiment['id'],
            'mask_path': mask_path,
            'mask_dict': mask_dict
            }

    with open(dict_path, 'w') as f:
        json.dump(full_dict, f, indent=2)

    experiment['nice_dict_path'] = dict_path

    return experiment


def row_col_from_roi(roi):
    """from an roi dict, get a list of rows and columns
    that are True in the mask_matrix

    Parameters
    ----------
    roi : dict
        x: x (col) offset
        y: y (row) offset
        width: col width
        height: row height
        mask_matrix: nested boolean list h x w

    Returns
    -------
    masked : :class:`numpy.ndarry`
        N x 2 array of [row, col] pairs

    """
    x0 = roi["x"]
    y0 = roi["y"]
    w = roi["width"]
    h = roi["height"]
    coord = np.mgrid[y0:(y0 + h):1, x0:(x0 + w):1]
    mask = np.array(roi['mask_matrix'])
    masked = np.array([coord[i][mask]
                       for i in range(coord.shape[0])]).reshape(2, -1).T
    return masked


def layered_mask_from_rois(rois: List[Any],
                           shape: Tuple[int, int]) -> np.ndarray:
    """makes a (nz, *shape) dimensioned array where each layer ("z")
    has no overlapping ROIs

    Parameters
    ----------
    rois: list
        list of ROIs. The list is the same as "cell_rois" in
        ExperimentSchema
    shape: tuple
        (nrows, ncols) of the destination image

    Returns
    -------
    masks: numpy.ndarray
       dimensions (nz, nrows, ncols)

    """
    def new_layer():
        return np.zeros(shape, dtype=np.uint32)

    masks = [new_layer()]
    for roi in rois:
        tmp = new_layer()
        tmp[roi['y']:(roi['y'] + roi['height']),
            roi['x']:(roi['x'] + roi['width'])] = np.uint32(roi['mask_matrix'])
        tmp *= roi['id']
        bool_tmp = tmp != 0
        placed = False
        for imask in range(len(masks)):
            bool_mask = masks[imask] != 0
            if np.any(bool_tmp & bool_mask):
                # some overlap, move to next layer
                pass
            else:
                # no overlap, add to this layer
                masks[imask] += tmp
                placed = True
                break
        if not placed:
            masks.append(tmp)
    masks = np.array(masks)

    return masks


def labeled_mask_from_experiment(exp):
    """create a labeled mask and to-cell-id dictionary
    from the experiment json

    Parameters
    ----------
    exp : dict
        id : unique experiment id from LIMS
        ophys_average_intensity_projection_image : path to
        intensity image. used to establish lateral size
        of the mask
        cell_rois : list of dict
            x: x (col) offset
            y: y (row) offset
            width: col width
            height: row height
            mask_matrix: nested boolean list h x w
            id : unique cell id, uint32 from LIMS

    Returns
    -------
    relabeled : :class:`numpy.ndarray`
        depth x n x m uint16 mask with new labels
    rdict : dict
        str(label in mask): cell ID

    """
    with PIL.Image.open(
            exp['ophys_average_intensity_projection_image']) as fim:
        im = np.array(fim)
    # layered mask with uint32 ids as intensities
    mask = layered_mask_from_rois(exp['cell_rois'], im.shape)

    # relabel as ordered uint16 and a translating map
    x = np.unique(mask)
    relabeled = np.zeros_like(mask).astype('uint16')
    rdict = {}
    for i, ix in enumerate(x):
        ind = np.nonzero(mask == ix)
        relabeled[ind] = i
        rdict[str(i)] = int(ix)

    # do not list the background
    rdict.pop("0")

    return relabeled, rdict


def calc_first_order_properties(M):
    """human-readable affine properties of a transform matrix

    Parameters
    ----------
    M : :class:`numpy.ndarray`
       3 x 3 augmented affine or projective matrix

    Returns
    -------
    summary : dict
        human-readable dict of transform properties

    """

    summary = {}
    if not np.all(M[2] == np.array([0.0, 0.0, 1.0])):
        summary['warning'] = "affine properties of projective transform"

    tform = AffineTransform(M)
    summary['scale'] = tform.scale
    summary['shear'] = tform.shear
    summary['translation'] = tuple(tform.translation)
    summary['rotation'] = tform.rotation

    return summary


def read_tiff_3d(filename):
    """read in 3D tiff for mask

    Parameters
    ----------
    filename : str
        path to tiff file

    Returns
    -------
    img : :class:`numpy.ndarray`
        depth x n x m image
    """

    with PIL.Image.open(filename) as ims:
        sz = ims.size
        nframes = ims.n_frames
        im = np.zeros((nframes, sz[1], sz[0])).astype('uint16')
        for i, iim in enumerate(PIL.ImageSequence.Iterator(ims)):
            im[i] = np.array(iim)

    return im



import cv2
import numpy as np


def contrast_adjust(image, CLAHE_grid, CLAHE_clip):
    """contrast adjust images

    Parameters
    ----------
    image: numpy.ndaray
        image to pre-process. CV_8UC1 (uint8) or CV_16UC1 (uint16)
    CLAHE_grid : int
        passed as tileGridSize to cv2.createCLAHE. If -1, skipped
    CLAHE_clip : float
        passed as clipLimit to cv2.createCLAHE

    Returns
    -------
    image: numpy.narray
        contrast-adjusted image

    """
    if CLAHE_grid != -1:
        clahe = cv2.createCLAHE(
            clipLimit=CLAHE_clip,
            tileGridSize=(CLAHE_grid, CLAHE_grid))
        image = clahe.apply(src=image)
    return image


def warp_image(image, tform, motion_type, dst_shape):
    """

    Parameters
    ----------
    image: numpy.ndarray
        image to warp. Assumes uint8
    tform: numpy.ndarray
        3x3 transformation matrix
    motion_type: str
        kind of motion. See function `register_image_pair` for possible
        values
    dst_shape: tuple
        desired shape of warped image

    Returns
    -------
    image: numpy.ndarray
        warped image, forced to uint8

    """
    if motion_type == 'MOTION_HOMOGRAPHY':
        warp = cv2.warpPerspective
    else:
        warp = cv2.warpAffine
        if tform.shape[0] == 3:
            tform = tform[0:2]

    image = warp(
            src=image,
            M=tform,
            dsize=dst_shape[::-1],
            flags=cv2.INTER_LINEAR + cv2.WARP_INVERSE_MAP).astype(np.uint8)

    return image



import os
import numpy as np
import pandas as pd
import json
import PIL.Image
import cv2
import scipy.spatial
import scipy.optimize
import networkx as nx
import itertools
import datetime
from pathlib import Path
from argschema import ArgSchemaParser

from nway.schemas import PairwiseMatchingSchema, PairwiseOutputSchema
from nway.meta_registration import MetaRegistration
import nway.utils as utils
import nway.image_processing_utils as imutils


class PairwiseException(Exception):
    pass


def gen_assignment_pairs(cost_matrix, solver):
    """generate pairs via an assignment problem solver

    Parameters
    ----------
    cost_matrix : :class:`pandas.DataFrame`
        index = labels1, columns = labels2
        N x M calculated cost matrix
    solver : str
       one of "Hungarian" (scipy), or "Blossom"

    Returns
    -------
    pairlabels : list
        pairs of labels [label from row (index), label from column]
        that are the assigned matches

    """

    cost_matrix_array = np.array(cost_matrix)

    if solver == 'Hungarian':
        assigned_pairs = np.transpose(
                np.array(
                    scipy.optimize.linear_sum_assignment(
                        cost_matrix_array)))
        row_lab = cost_matrix.index.tolist()
        col_lab = cost_matrix.columns.tolist()
        pairlabels = [[row_lab[pair[0]], col_lab[pair[1]]]
                      for pair in assigned_pairs]

    elif solver == 'Blossom':
        nrow, ncol = cost_matrix_array.shape
        # Blossom method is max weight, Hungarian is min weight
        weight = 1.0 / (1.0 + cost_matrix_array)
        ind = np.argwhere(weight > 0.002)
        etuples = []
        for i in ind:
            etuples.append((
                cost_matrix.index[i[0]],
                cost_matrix.columns[i[1]],
                {'weight': weight[i[0], i[1]]}))

        G = nx.Graph()
        G.add_edges_from(etuples)
        nxresults = nx.max_weight_matching(G)
        pairlabels = list([list(i) for i in nxresults])

        # sometimes networkx reverses the order
        for i in range(len(pairlabels)):
            if (
                    (pairlabels[i][0] in cost_matrix.index) &
                    (pairlabels[i][1] in cost_matrix.columns)):
                continue
            else:
                pairlabels[i] = pairlabels[i][::-1]

    return pairlabels


def gen_matching_table(
        cost_matrix, assigned_pairs, distance, iou, max_distance):
    """generate rich summary of the matched pairs

    Parameters
    ----------
    cost_matrix : :class:`pandas.DataFrame`
        index = labels1, columns = labels2
        N x M calculated cost matrix
    assigned_pairs : list
        pairs of labels [label from row (index), label from column]
        that are the assigned matches
    distance : :class:`pandas.DataFrame`
        index = labels1, columns = labels2
        Euclidean distance of centers
    iou : :class:`pandas.DataFrame`
        index = labels1, columns = labels2
        intersection-over-union pixel area ratios
    maximum_distance : Int
        threshold beyond which a large cost
        will be incurred

    Returns
    -------
    matches : list of dict
        list of matches
        each match:
        {
          "distance": euclidean distance between centers
          "iou": intersection over union of mask pixels
          "fixed": fixed cell id
          "moving": moving cell id
          "cost": calculated cost
        }
    rejected : list of dict
        list of pairs within max_dist threshold which were not chosen
        same structure as matches

    """

    # candidates within max_distance, but not assigned
    rejected = []
    matches = []
    cols = cost_matrix.columns.tolist()
    rows = cost_matrix.index.tolist()
    for row, col in itertools.product(rows, cols):
        if distance[col][row] < max_distance:
            imatch = {
                    "fixed": row,
                    "moving": col,
                    'distance': distance[col][row],
                    'iou': iou[col][row],
                    'cost': cost_matrix[col][row]}
            if [row, col] in assigned_pairs:
                matches.append(imatch)
            else:
                rejected.append(imatch)

    return matches, rejected


def region_properties(mask, mdict):
    """Characterize each label region in a mask

    Parameters
    ----------
    mask : :class:`numpy.ndarray`
        3D mask image (n x row x column)
    mdict : dict
        mapping for mask
        {
          "experiment": id of experiment
          "mask_path": absolute path to mask
          "mask_dict": {str(intensity): cellid}
        }

    Returns
    -------
    prop : dict
        'centers' : :class:`numpy.ndarray`
            (N x 2) Cartesian coordinates of region
            centers-of-mass
        'pixels' : set(tuples)
            (row, column) tuples of pixels for each label

    """
    prop = {}
    n = mask.max()
    prop['centers'] = np.zeros((n, 2))
    prop['pixels'] = [[]] * n
    prop['labels'] = [[]] * n
    for i in np.arange(n):
        label_locs = np.argwhere(mask == i + 1)
        prop['labels'][i] = mdict['mask_dict'][str(i + 1)]
        if len(label_locs) > 0:
            prop['centers'][i] = label_locs.mean(axis=0)[[-1, -2]]

        prop['pixels'][i] = set([tuple(x) for x in label_locs])

    # sometimes a cell ROI gets transformed outside
    # add a null column for it
    for label in mdict['mask_dict'].values():
        if label not in prop['labels']:
            prop['labels'].append(label)
            prop['centers'] = np.append(
                    prop['centers'],
                    np.array([0, 0]).reshape(1, -1), axis=0)
            prop['pixels'].append(set())
    return prop


def calculate_distance_and_iou(mask1, mask2, dict1, dict2):
    """Compute distance and intersection-over-union
    for the labels in two masks

    Parameters
    ----------
    mask1 : :class:`numpy.ndarray`
        3D mask image (n x row x column)
    mask2 : :class:`numpy.ndarray`
        3D mask image (n x row x column)
    dict1 : dict
        mapping for mask1
        {
          "experiment": id of experiment
          "mask_path": absolute path to mask
          "mask_dict": {str(intensity): cellid}
        }
    dict2 : dict
        mapping for mask2

    Returns
    -------
    distance : :class:`pandas.DataFrame`
        index = labels1, columns = labels2
        Euclidean distance of centers
    iou : :class:`pandas.DataFrame`
        index = labels1, columns = labels2
        intersection-over-union pixel area ratios

    """
    prop1 = region_properties(mask1, dict1)
    prop2 = region_properties(mask2, dict2)

    # pair-wise distances between cell centers
    distance = scipy.spatial.distance.cdist(
            prop1['centers'], prop2['centers'], 'euclidean')

    # intersection-over-union
    iou = np.zeros_like(distance)
    for i, ipix in enumerate(prop1['pixels']):
        for j, jpix in enumerate(prop2['pixels']):
            intersection = len(ipix.intersection(jpix))
            union = len(ipix.union(jpix))
            iou[i, j] = float(intersection) / union

    distance = pd.DataFrame(distance, prop1['labels'], prop2['labels'])
    iou = pd.DataFrame(iou, prop1['labels'], prop2['labels'])

    return distance, iou


def calculate_cost_matrix(distance, iou, maximum_distance):
    """Combine distance and iou into a cost

    Parameters
    ----------
    distance : :class:`pandas.DataFrame`
        index = labels1, columns = labels2
        Euclidean distance of centers
    iou : :class:`pandas.DataFrame`
        index = labels1, columns = labels2
        intersection-over-union pixel area ratios
    maximum_distance : Int
        threshold beyond which a large cost
        will be incurred

    Returns
    -------
    cost_matrix : :class:`pandas.DataFrame`
        index = labels1, columns = labels2
        N x M calculated cost matrix

    """
    norm_dist = np.array(distance) / maximum_distance
    # NOTE for the Hungarian method, MAX_COST is high to be ignored
    # relative to the typical cost range [0, 2]. For the Blossom method
    # high costs will not populate the graph as possible edges
    MAX_COST = 999.0
    norm_dist[norm_dist > 1] = MAX_COST
    cost_matrix = norm_dist + (1.0 - np.array(iou))
    cost_matrix = pd.DataFrame(
            cost_matrix, distance.index.tolist(), distance.columns.tolist())
    return cost_matrix


def cell_matching(mask1, mask2, dict1, dict2, max_dist, solver):
    """matching between 2 masks

    Parameters
    ----------
    mask1 : :class:`numpy.ndarray`
        3D mask image (n x row x column)
    mask2 : :class:`numpy.ndarray`
        3D mask image (n x row x column)
    dict1 : dict
        mapping for mask1
        {
          "experiment": id of experiment
          "mask_path": absolute path to mask
          "mask_dict": {str(intensity): cellid}
        }
    dict2 : dict
        mapping for mask2
    max_dist : int
        maximum distance threshold [pixels]. In the case of
        the Hungarian minimum weight algorithm, pairs beyond this
        threshold will receive a very large weight. In the case
        of the Blossom algorithm, pairs beyond this threshold will
        not be part of the graph.
    solver : str
       one of "Hungarian-cpp", "Hungarian" (scipy), or "Blossom"

    Returns
    -------
    cost_matrix : :class:`pandas.DataFrame`
        index = labels1, columns = labels2
        N x M calculated cost matrix
    matches : list of dict
        list of matches
        each match:
        {
          "distance": euclidean distance between centers
          "iou": intersection over union of mask pixels
          "fixed": fixed cell id
          "moving": moving cell id
          "cost": calculated cost
        }
    rejected : list of dict
        list of pairs within max_dist threshold which were not chosen
        same structure as matches
    unmatched : dict
        list of cell IDs from fixed and moving experiments that were
        never considered for a pair

    """

    distance, iou = calculate_distance_and_iou(mask1, mask2, dict1, dict2)

    cost_matrix = calculate_cost_matrix(distance, iou, max_dist)

    assigned_pairs = gen_assignment_pairs(cost_matrix, solver)

    matches, rejected = gen_matching_table(
            cost_matrix,
            assigned_pairs,
            distance,
            iou,
            max_dist)

    # catalog what did not get paired at all
    paired_fixed_ids = [i['fixed'] for i in matches] + \
        [i['fixed'] for i in rejected]
    paired_moving_ids = [i['moving'] for i in matches] + \
        [i['moving'] for i in rejected]
    unmatched = {
            'fixed': [i for i in dict1['mask_dict'].values()
                      if i not in paired_fixed_ids],
            'moving': [i for i in dict2['mask_dict'].values()
                       if i not in paired_moving_ids]
            }

    return cost_matrix, matches, rejected, unmatched


def transform_mask(moving, dst_shape, tform):
    """warp a mask according to a transform

    Parameters
    ----------
    moving : :class:`numpy.ndarray`
        nlayers x n x m int
    dst_shape : tuple
        (n', m') shape of destination layers
    tform : :class:`numpy.ndarry`
        3 x 3 transformation matrix

    Returns
    -------
    transformed_3d : :class:`numpy.ndarray`
        nlayer x n' x m' int

    """

    transformed_3d = np.zeros(
            (
                moving.shape[0],
                dst_shape[0],
                dst_shape[1]),
            dtype=np.int)

    for k, frame in enumerate(moving):
        labels = np.unique(frame)
        transformed_2d = np.zeros(dst_shape)

        for label in labels:
            tmp = np.zeros_like(frame).astype('float32')
            tmp[frame == label] = 1
            tmp_registered = cv2.warpPerspective(
                    tmp,
                    tform,
                    dst_shape[::-1],
                    flags=cv2.INTER_LINEAR + cv2.WARP_INVERSE_MAP)

            transformed_2d[tmp_registered > 0] = label

        transformed_3d[k, :, :] = transformed_2d

    return transformed_3d


class PairwiseMatching(ArgSchemaParser):
    default_schema = PairwiseMatchingSchema
    default_output_schema = PairwiseOutputSchema
    """Class for matching cells between optical physiology sessions
    """

    def run(self):
        """main function call for PairwiseMatching
        """
        self.logger.name = type(self).__name__
        logging_prefix = "Matching {} to {}".format(
                self.args['fixed']['id'],
                self.args['moving']['id'])
        self.logger.info(logging_prefix)

        # identifying strings for filenames
        fixed_strid = '%s' % self.args['fixed']['id']
        moving_strid = '%s' % self.args['moving']['id']

        # register the average intensity images
        fixp = self.args['fixed']['ophys_average_intensity_projection_image']
        movp = self.args['moving']['ophys_average_intensity_projection_image']
        with PIL.Image.open(fixp) as im:
            img_fixed = np.array(im)
        with PIL.Image.open(movp) as im:
            img_moving = np.array(im)

        meta_reg = MetaRegistration(
                maxCount=self.args['registration_iterations'],
                epsilon=self.args['registration_precision'],
                motion_type=self.args['motionType'],
                gaussFiltSize=self.args['gaussFiltSize'],
                CLAHE_grid=self.args['CLAHE_grid'],
                CLAHE_clip=self.args['CLAHE_clip'],
                edge_buffer=self.args['edge_buffer'],
                include_original=self.args['include_original'])
        meta_reg(img_moving, img_fixed)
        self.logger.info(f"{logging_prefix}: best registration was "
                         f"{meta_reg.best_candidate}")

        self.tform = meta_reg.best_matrix

        if meta_reg.best_candidate == ["Identity"]:
            self.tform = self.tform.astype('float')
            output_json = {
                    'unmatched': {},
                    'matches': [],
                    'rejected': [],
                    'fixed_experiment': self.args['fixed']['id'],
                    'moving_experiment': self.args['moving']['id'],
                    'transform': {
                        "best_registration": meta_reg.best_candidate,
                        "properties": utils.calc_first_order_properties(
                            self.tform),
                        'matrix': self.tform.tolist(),
                        'transform_type': "IDENTITY"
                        }
                    }
            self.output(output_json, indent=2)
            fname = "avg_projections_{}_{}_{}.png".format(
                self.args['fixed']['id'], self.args['moving']['id'],
                datetime.datetime.now().strftime("%Y%m%d%H%M%S"))
            fpath = utils.write_pair_images(
                        Path(self.args['output_directory']) / fname,
                        titles=[
                            self.args['fixed']['id'],
                            self.args['moving']['id']],
                        images=[img_fixed, img_moving],
                        suptitle=("failed to find registration for these "
                                  "average projection pairs."))
            self.logger.warn(f"{logging_prefix}: no registration found. "
                             f"wrote {fpath}")
            return

        if self.args['save_registered_image']:
            moving_warped = imutils.warp_image(img_moving, self.tform,
                                               self.args['motionType'],
                                               img_fixed.shape)
            imfname = os.path.join(
                self.args['output_directory'],
                'register_%s_to_%s.tif' % (moving_strid, fixed_strid))
            with PIL.Image.fromarray(moving_warped) as im:
                im.save(imfname)

        # transform moving segmentation mask
        segmask_fixed_3d = utils.read_tiff_3d(
                self.args['fixed']['nice_mask_path'])
        segmask_moving_3d = utils.read_tiff_3d(
                self.args['moving']['nice_mask_path'])
        segmask_moving_3d_registered = transform_mask(
                segmask_moving_3d,
                segmask_fixed_3d.shape[1:],
                self.tform)

        with open(self.args['fixed']['nice_dict_path'], 'r') as f:
            fixed_dict = json.load(f)
        with open(self.args['moving']['nice_dict_path'], 'r') as f:
            moving_dict = json.load(f)

        # matching cells
        self.cost_matrix, self.matches, self.rejected, self.unmatched = \
            cell_matching(
                    segmask_fixed_3d,
                    segmask_moving_3d_registered,
                    fixed_dict,
                    moving_dict,
                    self.args['maximum_distance'],
                    self.args['assignment_solver'])

        # opencv likes float32, but json does not
        self.tform = self.tform.astype('float')
        output_json = {
                'unmatched': self.unmatched,
                'matches': self.matches,
                'rejected': self.rejected,
                'fixed_experiment': self.args['fixed']['id'],
                'moving_experiment': self.args['moving']['id'],
                'transform': {
                    "best_registration": meta_reg.best_candidate,
                    "properties": utils.calc_first_order_properties(
                        self.tform),
                    'matrix': self.tform.tolist(),
                    'transform_type': self.args['motionType']
                    }
                }
        self.output(output_json, indent=2)

        return


if __name__ == "__main__":  # pragma: no cover
    pmod = PairwiseMatching()
    pmod.run()



from argschema.schemas import DefaultSchema
from argschema import ArgSchema
from argschema.fields import (
        Boolean, Int, Str, Float, Dict,
        List, InputFile, OutputDir, Nested,
        OutputFile, Bool)
import marshmallow as mm
import logging
import json

logger = logging.getLogger(__name__)


class ExperimentSchema(DefaultSchema):
    id = Int(
        required=True,
        description="experiment id")
    ophys_average_intensity_projection_image = InputFile(
        required=True,
        description="max projection intensity image")
    stimulus_name = Str(
        required=False,
        allow_none=True,
        description="Name of stimulus for a given experiment")
    cell_rois = List(
        Dict,
        required=True,
        cli_as_single_argument=True,
        description='dict mapping of ids, labels, zs,')
    nice_mask_path = InputFile(
        required=True,
        description="path to mask tiff with unique labels")
    nice_dict_path = InputFile(
        required=True,
        description="path to dict for mask labels to LIMS ids")


class CommonMatchingSchema(ArgSchema):
    log_level = Str(
        default="INFO",
        description="override argschema default")
    maximum_distance = Int(
        required=False,
        default=10,
        description=("Maximum distance (in pixels) between two cells,"
                     " above which a match is always rejected."))
    registration_iterations = Int(
        required=False,
        default=1000,
        description=("Number of iterations for intensity-"
                     "based registration"))
    registration_precision = Float(
        required=False,
        default=1.5e-7,
        description=("Threshold of squared error, below which "
                     "registration is terminated"))
    assignment_solver = Str(
        required=False,
        default="Blossom",
        missing="Blossom",
        validator=mm.validate.OneOf([
            "Blossom",
            "Hungarian"]),
        description=("What method to use for solving the assignment problem"
                     " in pairwise matching"))
    motionType = Str(
        required=False,
        missing="MOTION_EUCLIDEAN",
        default="MOTION_EUCLIDEAN",
        validator=mm.validate.OneOf([
            "MOTION_TRANSLATION",
            "MOTION_EUCLIDEAN",
            "MOTION_AFFINE",
            "MOTION_HOMOGRAPHY"
            ]))
    gaussFiltSize = Int(
        required=False,
        missing=5,
        default=5,
        description="passed to opencv findTransformECC")
    CLAHE_grid = Int(
        required=False,
        default=24,
        missing=24,
        description="tileGridSize for cv2 CLAHE, set to -1 to disable CLAHE")
    CLAHE_clip = Float(
        required=False,
        default=2.5,
        missing=2.5,
        description="clipLimit for cv2 CLAHE")
    edge_buffer = Int(
        required=False,
        default=40,
        missing=40,
        description=("in the Crop transform for meta-registration, this "
                     "many pixels will be cropped from the top, bottom, "
                     "left, and right of both images."))
    include_original = Bool(
        required=False,
        default=False,
        missing=False,
        description=("whether to include the original registration strategy "
                     "which occasionally succeeds with a not great result."))

    @mm.post_load
    def hungarian_warn(self, data, **kwargs):
        if "Hungarian" in data['assignment_solver']:
            logger.warning("Hungarian method not recommended. It is not "
                           "stable under permutations for typical cell "
                           "matching. Use Blossom.")
        return data


class NwayMatchingSchema(CommonMatchingSchema):
    ''' Class that uses argschema to take care of input arguments '''
    output_directory = OutputDir(
        required=True,
        description=("destination for output files. If None, will be set from "
                     "output_directory field in input file"))
    experiment_containers = Dict(
        required=True,
        description="contains data for matching")
    save_pairwise_results = Boolean(
        required=False,
        default=False,
        description=("Whether to save pairwise output jsons"))
    pruning_method = Str(
        required=False,
        missing="keepmin",
        default="keepmin",
        description=("method for reducing pruning graph: "
                     "'keepmin': similar to original, delete from "
                     "    graph neighbors of lowest subgraph score."
                     "'popmax': delete highest scores of subgraphs "
                     "    recursively"))
    parallel_workers = Int(
        required=False,
        missing=3,
        default=3,
        description=("number of parallel workers for multiprocessing pool "
                     "for executing pairwise matching. If 1, runs serially."))


class PairwiseMatchingSchema(CommonMatchingSchema):
    output_directory = OutputDir(
        required=True,
        description="destination for output files.")
    save_registered_image = Boolean(
        required=False,
        default=False,
        description='Whether to save registered image.')
    fixed = Nested(ExperimentSchema)
    moving = Nested(ExperimentSchema)


class NwayMatchingOutputSchema(DefaultSchema):
    nway_matches = List(
        List(Int),
        required=True,
        description="list of lists of matching IDs")
    pairwise_results = List(
        Dict,
        required=True,
        description="list of pairwise result dicts")
    nway_match_fraction_plot = OutputFile(
        required=True,
        description="Path of match fraction plot *.png")
    nway_warp_overlay_plot = OutputFile(
        required=True,
        description="Path of warp overlay plot *.png")
    nway_warp_summary_plot = OutputFile(
        required=True,
        description="Path of warp summary plot *.png")


class PairwiseOutputSchema(DefaultSchema):
    fixed_experiment = Int(
        required=True,
        description="fixed experiment id of pair")
    moving_experiment = Int(
        required=True,
        description="moving experiment id of pair")
    transform = Dict(
        required=True,
        description="transform applied to moving")
    matches = List(
        Dict,
        required=True,
        description="matches made from pairwise matching")
    rejected = List(
        Dict,
        required=True,
        description="pairs within max_distance, but not matched")
    unmatched = Dict(
        required=True,
        description="list of cell IDs that were never in a considered pair")


class NwayDiagnosticSchema(ArgSchema):
    output_pdf = OutputFile(
        required=True,
        description="path to output pdf")
    use_input_dir = Bool(
        required=False,
        missing=False,
        default=False,
        descriptip="output to same directory as input")


class NwayMatchSummarySchema(ArgSchema):
    input_file = InputFile(
        required=False,
        desc="Input *.json file path to nway matching."
    )
    output_file = InputFile(
        required=False,
        desc="Output *.json file from nway matching."
    )
    nway_input = Dict(
        required=True,
        desc="Input to nway matching in Python dictionary form."
    )
    nway_output = Dict(
        required=True,
        desc="Output of nway matching in Python dictionary form."
    )
    output_directory = OutputDir(
        required=True,
        description="Destination for summary plot output file(s).")

    @mm.pre_load
    def fill_dict_inputs(self, data: dict, **kwargs) -> dict:
        if not data['nway_input']:
            with open(data['input_file'], 'r') as f:
                input_dict = json.load(f)
            data['nway_input'] = input_dict
        elif data.get('input_file'):
            logger.warning("Both --nway_input and --input_file were provided "
                           "so --input_file will be ignored.")

        if not data['nway_output']:
            with open(data['output_file'], 'r') as f:
                output_dict = json.load(f)
            data['nway_output'] = output_dict
        elif data.get('output_file'):
            logger.warning("Both --nway_output and --output_file were "
                           "provided so --output_file will be ignored.")

        return data



import pytest
import contextlib
import cv2
import PIL.Image
import numpy as np
from pathlib import Path
import nway.transforms as nwtf
import nway.image_processing_utils as impu

TEST_FILE_DIR = Path.resolve(Path(__file__)).parent / "test_files"


@pytest.fixture
def fixed_image_fixture():
    # grab one specific file because failing large translation
    # cases vary image-by-image
    fgen = TEST_FILE_DIR.rglob("avgInt_a1X.png")
    for f in fgen:
        if "ophys_session_775289198" in str(f):
            impath = f
    with PIL.Image.open(impath) as fp:
        image = np.array(fp)
    return image


@pytest.fixture
def moving_image_fixture(fixed_image_fixture, request):
    immoving = cv2.warpPerspective(
            fixed_image_fixture,
            request.param.get("transform"),
            fixed_image_fixture.shape[::-1])
    return immoving, request.param


@pytest.mark.parametrize(
        "src, dst",
        [
            (
                np.uint8([[1, 2, 3, 4],
                          [4, 1, 2, 3],
                          [3, 4, 1, 2],
                          [2, 3, 4, 1]]),
                np.uint8([[5, 6, 7, 8],
                          [8, 5, 6, 7],
                          [7, 8, 5, 6],
                          [6, 7, 8, 5]]))])
def test_Transform(src, dst):
    tf = nwtf.Transform()
    tf.estimate(src, dst)
    np.testing.assert_array_equal(tf.matrix, np.eye(3).astype('float32'))
    mov, fix = tf.apply(src, dst)
    np.testing.assert_array_equal(src, mov)
    np.testing.assert_array_equal(fix, dst)


@pytest.mark.parametrize(
        "src, dst, grid, clip",
        [
            (
                np.random.randint(0, 255, dtype='uint8', size=(482, 512)),
                np.random.randint(0, 255, dtype='uint8', size=(482, 512)),
                24, 2.5
                )])
def test_CLAHE(src, dst, grid, clip):
    tf = nwtf.CLAHE(CLAHE_grid=grid, CLAHE_clip=clip)
    assert tf.CLAHE_grid == grid
    assert tf.CLAHE_clip == clip
    # no geometrical transform here, should inherit identity
    tf.estimate(src, dst)
    np.testing.assert_array_equal(tf.matrix, np.eye(3).astype('float32'))
    # just check shapes
    mov, fix = tf.apply(src, dst)
    assert mov.shape == src.shape
    assert fix.shape == dst.shape


@pytest.mark.parametrize(
        "src, expected",
        [
            (
                np.uint8([[0, 0, 0, 0, 0, 0, 0, 0],
                          [0, 0, 0, 0, 0, 0, 0, 0],
                          [0, 0, 4, 4, 4, 0, 0, 0],
                          [0, 0, 8, 9, 8, 0, 0, 0],
                          [0, 0, 3, 3, 3, 0, 0, 0],
                          [0, 0, 0, 0, 0, 0, 0, 0],
                          [0, 0, 0, 0, 0, 0, 0, 0],
                          [0, 0, 0, 0, 0, 0, 0, 0]]),
                np.float32([[1.0, 0.0, -1.0],
                            [0.0, 1.0, 2.0],
                            [0.0, 0.0, 1.0]]))])
def test_PhaseCorrelate(src, expected):
    dst = impu.warp_image(src, expected, "MOTION_EUCLIDEAN", src.shape)
    tf = nwtf.PhaseCorrelate()
    tf.estimate(src, dst)
    mov, fix = tf.apply(src, dst)
    np.testing.assert_array_equal(fix, dst)
    np.testing.assert_array_equal(mov, dst)
    np.testing.assert_allclose(tf.matrix, expected, rtol=1e-3)


@pytest.mark.parametrize(
        "src, dst, edge, expected_src, expected_dst, context",
        [
            (
                np.uint8([[1, 2, 3, 4],
                          [0, 1, 2, 3],
                          [2, 3, 4, 4],
                          [2, 3, 6, 7]]),
                np.uint8([[1, 2, 3, 4],
                          [0, 5, 7, 3],
                          [2, 5, 8, 4],
                          [2, 3, 6, 7]]),
                1,
                np.uint8([[1, 2],
                          [3, 4]]),
                np.uint8([[5, 7],
                          [5, 8]]),
                contextlib.nullcontext()),
            (
                np.uint8([[1, 2, 3, 4],
                          [0, 1, 2, 3],
                          [2, 3, 4, 4],
                          [2, 3, 6, 7]]),
                np.uint8([[1, 2, 3, 4],
                          [0, 5, 7, 3],
                          [2, 5, 8, 4],
                          [2, 3, 6, 7]]),
                0,
                np.uint8([[1, 2, 3, 4],
                          [0, 1, 2, 3],
                          [2, 3, 4, 4],
                          [2, 3, 6, 7]]),
                np.uint8([[1, 2, 3, 4],
                          [0, 5, 7, 3],
                          [2, 5, 8, 4],
                          [2, 3, 6, 7]]),
                contextlib.nullcontext()),
            (
                np.uint8([[1, 2, 3, 4],
                          [0, 1, 2, 3],
                          [2, 3, 4, 4],
                          [2, 3, 6, 7]]),
                np.uint8([[1, 2, 3, 4],
                          [0, 5, 7, 3],
                          [2, 5, 8, 4],
                          [2, 3, 6, 7]]),
                3,
                np.uint8([[1, 2, 3, 4],
                          [0, 1, 2, 3],
                          [2, 3, 4, 4],
                          [2, 3, 6, 7]]),
                np.uint8([[1, 2, 3, 4],
                          [0, 5, 7, 3],
                          [2, 5, 8, 4],
                          [2, 3, 6, 7]]),
                pytest.raises(nwtf.TransformException))
                ])
def test_Crop(src, dst, edge, expected_src, expected_dst, context):
    tf = nwtf.Crop(edge_buffer=edge)
    tf.estimate(src, dst)
    with context:
        csrc, cdst = tf.apply(src, dst)
        np.testing.assert_array_equal(expected_src, csrc)
        np.testing.assert_array_equal(expected_dst, cdst)


@pytest.mark.parametrize("motion_type, tform_tol, trans_tol", [
    ("MOTION_TRANSLATION", (0.0001, 0.0001), (0.5, 0.0)),
    ("MOTION_EUCLIDEAN", (0.0002, 0.0001), (0.5, 0.0)),
    ("MOTION_AFFINE", (0.0005, 0.005), (0.5, 0.0)),
    ("MOTION_HOMOGRAPHY", (0.005, 0.0001), (0.5, 0.0)),
    ])
@pytest.mark.parametrize(
        "moving_image_fixture",
        [
           # identity
           {"transform": np.array([[1.0, 0.0, 0.0],
                                   [0.0, 1.0, 0.0],
                                   [0.0, 0.0, 1.0]])},
           # x translation
           {"transform": np.array([[1.0, 0.0, 10.0],
                                   [0.0, 1.0, 0.0],
                                   [0.0, 0.0, 1.0]])},
           # x and y translation
           {"transform": np.array([[1.0, 0.0, 10.0],
                                   [0.0, 1.0, -23.0],
                                   [0.0, 0.0, 1.0]])},
           # 5 deg rotation and x and y translation
           {"transform": np.array([[np.cos(0.087), -np.sin(0.087), 10.0],
                                   [np.sin(0.087), np.cos(0.087), -23.0],
                                   [0.0, 0.0, 1.0]])},
           ],
        indirect=["moving_image_fixture"])
def test_ECC(fixed_image_fixture, moving_image_fixture,
             motion_type, tform_tol, trans_tol):
    moving_image, param = moving_image_fixture
    expected_transform = param["transform"]

    tf = nwtf.ECC(maxCount=500, epsilon=1e-3, motion_type=motion_type,
                  gaussFiltSize=5)
    tf.estimate(moving_image, fixed_image_fixture)
    transform = tf.matrix

    if not np.allclose(expected_transform[0:2, 0:2], np.eye(2)) & \
            (motion_type == 'MOTION_TRANSLATION'):
        # translation can't be expected to solve for rotation
        pass
    else:
        # the multiplicative part of the transform
        np.testing.assert_allclose(transform[0:2, 0:2],
                                   expected_transform[0:2, 0:2],
                                   atol=tform_tol[0], rtol=tform_tol[1])
        # the translation part of the transform
        np.testing.assert_allclose(transform[0:2, 2],
                                   expected_transform[0:2, 2],
                                   atol=trans_tol[0], rtol=trans_tol[1])

        mov, _ = tf.apply(moving_image, fixed_image_fixture)
        np.testing.assert_allclose(
                mov.astype(int),
                fixed_image_fixture.astype(int),
                atol=26)


def test_TransformList():
    # chain together 2 identities
    tf1 = nwtf.Transform()
    tf2 = nwtf.Transform()
    tf_list = nwtf.TransformList(transforms=[tf1, tf2])

    src = np.random.randint(0, 255, dtype='uint8', size=(512, 512))
    dst = np.random.randint(0, 255, dtype='uint8', size=(512, 512))

    tf_list.estimate(src, dst)
    np.testing.assert_array_equal(tf_list.matrix, np.eye(3).astype('float32'))

    mov, fix = tf_list.apply(src, dst)
    np.testing.assert_array_equal(src, mov)
    np.testing.assert_array_equal(fix, dst)



from setuptools import setup, find_packages


with open('requirements.txt', 'r') as f:
    required = f.read().splitlines()


setup(name="ophys_nway_matching",
      use_scm_version=True,
      description=("building up a repo from the ophys_ophys_nway "
                   "running in LIMS, Aug 2019"),
      author="Daniel Kapner, original code Fuhui Long",
      author_email='danielk@alleninstitute.org',
      url="https://github.com/AllenInstitute/ophys_nway_matching",
      packages=find_packages(),
      setup_requires=['setuptools_scm'],
      install_requires=required)



import json
import os
import shutil
import time

import matplotlib
import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt
import numpy as np
import nway.diagnostics as nwdi
import PIL.Image
import PyPDF2
import pytest

import pandas as pd

matplotlib.use('Agg')


TEST_FILE_DIR = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        'test_files')


@pytest.fixture(scope='module')
def new_output():
    ojpath = os.path.join(
            TEST_FILE_DIR,
            "test0",
            "new_output.json")
    yield ojpath


def test_cell_lookup(new_output):
    cell_lookup = nwdi.cell_experiment_dict(new_output)
    cells = list(cell_lookup.keys())
    with open(new_output, 'r') as f:
        outj = json.load(f)
    for match in outj['nway_matches']:
        for cell_id in match:
            assert cell_id in cells


def test_pairwise_transforms(new_output):
    ptf = nwdi.pairwise_transforms(new_output)
    assert type(ptf) == dict

    fig = plt.figure(1)
    ptf = nwdi.pairwise_transforms(new_output, fig=fig)
    assert type(ptf) == dict

    outer_plot_grid = gridspec.GridSpec(2, 2, wspace=0.2, hspace=0.2)
    subplot_spec = outer_plot_grid[1]
    fig.clf()
    ptf = nwdi.pairwise_transforms(
            new_output, fig=fig, subplot_spec=subplot_spec)
    assert type(ptf) == dict


def test_some_grid():
    for n in range(1, 20):
        (nr, nc) = nwdi.some_grid(n)
        assert (nr * nc) >= n


def test_pairwise_matches(new_output):
    costs, allcosts = nwdi.pairwise_matches(new_output)
    assert type(costs) == dict
    assert type(allcosts) == dict

    fig = plt.figure(1)
    costs, allcosts = nwdi.pairwise_matches(new_output, fig=fig)
    assert type(costs) == dict
    assert type(allcosts) == dict

    outer_plot_grid = gridspec.GridSpec(2, 2, wspace=0.2, hspace=0.2)
    subplot_spec = outer_plot_grid[1]
    fig.clf()
    costs, allcosts = nwdi.pairwise_matches(
            new_output, fig=fig, subplot_spec=subplot_spec)
    assert type(costs) == dict
    assert type(allcosts) == dict


def test_nway_matches(new_output):
    allnw = nwdi.nway_matches(new_output)
    assert type(allnw) == dict

    fig = plt.figure(1)
    allnw = nwdi.nway_matches(new_output, fig=fig)
    assert type(allnw) == dict


def test_plot_all(new_output, tmpdir):
    nwdi.plot_all(new_output)

    output_dir = str(tmpdir.mkdir("diagnostic_test"))
    fname = os.path.join(
            output_dir,
            "output.pdf")

    nwdi.plot_all(new_output, fname)
    assert os.path.isfile(fname)
    PyPDF2.PdfFileReader(fname)


def test_NwayDiagnostics(new_output, tmpdir):
    output_dir = str(tmpdir.mkdir("diagnostic_test"))
    fname = os.path.join(
            output_dir,
            "output2.pdf")

    args = {
            'input_json': new_output,
            'output_pdf': fname,
            'log_level': "DEBUG"
            }

    nd = nwdi.NwayDiagnostics(input_data=args, args=[])
    nd.run()
    assert os.path.isfile(fname)
    PyPDF2.PdfFileReader(fname)

    ninput = os.path.join(
            output_dir,
            "tmpinput.json")

    fname3 = os.path.join(
            output_dir,
            "output3.pdf")

    shutil.copy(new_output, ninput)
    args = {
            'input_json': ninput,
            'output_pdf': fname3,
            'use_input_dir': True,
            'log_level': "DEBUG"
            }
    nd = nwdi.NwayDiagnostics(input_data=args, args=[])
    nd.run()

    oname = os.path.join(
            output_dir,
            os.path.basename(args['output_pdf']))
    assert os.path.isfile(oname)
    PyPDF2.PdfFileReader(oname)


@pytest.fixture
def nway_input_fixture(request, tmp_path):
    ophys_experiments = request.param

    expected_id_avg_image_map = {}
    for indx, expt in enumerate(ophys_experiments):
        timestamp = time.strftime('%Y_%m_%d_%H_%M_%S', time.localtime())
        save_fname = tmp_path / f"{indx}_{timestamp}.png"

        image_array = expt["dummy_avg_image"]
        expected_id_avg_image_map[expt['id']] = image_array

        with PIL.Image.fromarray(image_array) as im:
            im.save(save_fname)
        expt["ophys_average_intensity_projection_image"] = str(save_fname)

    nway_input_dict = {
        "experiment_containers": {
            "ophys_experiments": ophys_experiments
        }
    }
    return (nway_input_dict, expected_id_avg_image_map)


@pytest.mark.parametrize("nway_input_fixture, expected_id_stim_name_map", [
    # Test case 1
    ([{"id": 12345,
       "stimulus_name": "A",
       "dummy_avg_image": np.random.randint(0, 255, size=(5, 5),
                                            dtype=np.uint8)},
      {"id": 54321,
       "stimulus_name": "B",
       "dummy_avg_image": np.random.randint(0, 255, size=(5, 5),
                                            dtype=np.uint8)}],
     {12345: "A", 54321: "B"}),

    # Test case 2
    ([{"id": 56789,
       "stimulus_name": "B",
       "dummy_avg_image": np.random.randint(0, 255, size=(5, 5),
                                            dtype=np.uint8)},
      {"id": 98765,
       "stimulus_name": "A",
       "dummy_avg_image": np.random.randint(0, 255, size=(5, 5),
                                            dtype=np.uint8)}],
     {98765: "A", 56789: "B"}),

    # Test case 3
    ([{"id": 56789,
       "stimulus_name": None,
       "dummy_avg_image": np.random.randint(0, 255, size=(5, 5),
                                            dtype=np.uint8)},
      {"id": 98765,
       "stimulus_name": "B",
       "dummy_avg_image": np.random.randint(0, 255, size=(5, 5),
                                            dtype=np.uint8)}],
     {98765: "B", 56789: "Unknown Stimulus"})
], indirect=["nway_input_fixture"])
def test_create_nway_input_maps(nway_input_fixture, expected_id_stim_name_map):
    nway_input, expected_expt_id_avg_image_map = nway_input_fixture

    obt = nwdi.create_nway_input_maps(nway_input)

    assert obt[0] == expected_id_stim_name_map
    for k, v in expected_expt_id_avg_image_map.items():
        assert k in obt[1]
        assert np.allclose(v, obt[1][k])


@pytest.mark.parametrize(("nway_output, expt_id_stim_name_map, "
                          "expt_id_avg_image_map, expected_df, "
                          "expected_warped_images_map"), [
    # nway_output
    ({"pairwise_results": [
        {"fixed_experiment": 123,
         "moving_experiment": 321,
         "matches": [1, 2],
         "unmatched": {"fixed": [3], "moving": [4]},
         "transform": {"matrix": np.identity(3),
                       "transform_type": "IDENTITY"}},
        {"fixed_experiment": 123,
         "moving_experiment": 456,
         "matches": [1, 2, 3],
         "unmatched": {"fixed": [], "moving": [4]},
         "transform": {"matrix": np.identity(3),
                       "transform_type": "IDENTITY"}},
        {"fixed_experiment": 456,
         "moving_experiment": 321,
         "matches": [1, 2, 3, 4],
         "unmatched": {"fixed": [], "moving": []},
         "transform": {"matrix": np.identity(3),
                       "transform_type": "IDENTITY"}},
     ]},
     # expt_id_stim_name_map
     {123: "A", 321: "B", 456: "C"},
     # expt_id_avg_image_map
     {123: np.identity(9), 321: np.identity(9) * 2, 456: np.identity(9) * 3},
     # expected_df
     pd.DataFrame({
         "fixed_expt": [123, 123, 456],
         "fixed_expt_stim_name": ["A", "A", "C"],
         "moving_expt": [321, 456, 321],
         "moving_expt_stim_name": ["B", "C", "B"],
         "n_unmatched_fixed": [1, 0, 0],
         "n_unmatched_moving": [1, 1, 0],
         "n_matches": [2, 3, 4],
         "n_total": [4, 4, 4],
         "fraction_matched": [0.5, 3/4, 1]}),
     # expected_warped_images_map
     {"321_to_123": np.identity(9) * 2, "123_to_321": np.identity(9),
      "123_to_456": np.identity(9), "456_to_123": np.identity(9) * 3,
      "456_to_321": np.identity(9) * 3, "321_to_456": np.identity(9) * 2}),
])
def test_create_nway_summary_df(nway_output, expt_id_stim_name_map,
                                expt_id_avg_image_map, expected_df,
                                expected_warped_images_map):
    obt = nwdi.create_nway_summary_df(expt_id_stim_name_map,
                                      expt_id_avg_image_map,
                                      nway_output)

    pd.testing.assert_frame_equal(obt, expected_df)

    assert obt.attrs['expt_id_stim_name_map'] == expt_id_stim_name_map

    obt_warped_images = obt.attrs['warped_images']
    for k, v in expected_warped_images_map.items():
        assert k in obt_warped_images
        assert np.allclose(obt_warped_images[k], v)

    obt_avg_image_map = obt.attrs['expt_id_avg_image_map']
    for k, v in expt_id_avg_image_map.items():
        assert k in obt_avg_image_map
        assert np.allclose(obt_avg_image_map[k], v)


@pytest.fixture
def nway_summary_df_fixture() -> pd.DataFrame:

    random_image_array = np.random.randint(0, 255, size=(25, 25),
                                           dtype=np.uint8)

    summary_df = pd.DataFrame({
         "fixed_expt": [123, 123, 456],
         "fixed_expt_stim_name": ["A", "A", "C"],
         "moving_expt": [321, 456, 321],
         "moving_expt_stim_name": ["B", "C", "B"],
         "n_unmatched_fixed": [1, 0, 0],
         "n_unmatched_moving": [1, 1, 0],
         "n_matches": [2, 3, 4],
         "n_total": [4, 4, 4],
         "fraction_matched": [0.5, 3/4, 1]})
    summary_df.attrs["warped_images"] = {
        "321_to_123": random_image_array,
        "123_to_321": random_image_array,
        "123_to_456": random_image_array,
        "456_to_123": random_image_array,
        "456_to_321": random_image_array,
        "321_to_456": random_image_array}
    summary_df.attrs["expt_id_stim_name_map"] = {123: "A", 321: "B", 456: "C"}
    summary_df.attrs["expt_id_avg_image_map"] = {
        123: random_image_array,
        321: random_image_array,
        456: random_image_array}
    return summary_df


def test_plot_container_match_fraction(nway_summary_df_fixture):
    # Smoke test to ensure that *a* plot can be generated without errors
    obt = nwdi.plot_container_match_fraction(nway_summary_df_fixture)
    assert isinstance(obt, matplotlib.figure.Figure)


def test_plot_container_warp_overlays(nway_summary_df_fixture):
    # Smoke test to ensure that *a* plot can be generated without errors
    obt = nwdi.plot_container_warp_overlays(nway_summary_df_fixture)
    assert isinstance(obt, matplotlib.figure.Figure)


def test_plot_container_warp_summary(nway_summary_df_fixture):
    # Smoke test to ensure that *a* plot can be generated without errors
    obt = nwdi.plot_container_warp_summary(nway_summary_df_fixture)
    assert isinstance(obt, matplotlib.figure.Figure)



import numpy as np
import os
import json
import itertools
import copy
import multiprocessing
import pandas as pd
import networkx as nx
from nway.pairwise_matching import PairwiseMatching
from nway.schemas import NwayMatchingSchema, NwayMatchingOutputSchema
import nway.utils as utils
from nway import __version__ as nway_version
from argschema import ArgSchemaParser

from nway.diagnostics import NwaySummary


class NwayException(Exception):
    pass


def subgraphs(G):
    """returns connected component subgraphs. Replaces deprecated
    nx.connected_component_subgraphs()

    Parameters
    ----------
    G : :class:`networkx.Graph`

    Returns
    -------
    subgs : tuple
        tuple of connected subgraphs of G

    """
    subgs = (G.subgraph(c).copy() for c in nx.connected_components(G))
    return subgs


def greduce(glist, method='keepmin'):
    """recursively reduces a graph according to node attribute 'score'

    Parameters
    ----------
    glist : tuple
        tuple of subgraphs
    method : str
        'keepmin' reduce graph by finding the minimum score of a subraph
            and eliminating its neighbors
        'popmax' reduce graph by finding the maximum score and eliminating it

    Returns
    -------
    results : list
        list of nodes that survive the pruning

    """
    results = []
    for g in glist:
        gnodes = list(g.nodes.keys())
        if len(gnodes) == 1:
            results += [gnodes[0]]
        else:
            if method == 'popmax':
                maxind = np.argmax([i['score'] for i in g.nodes.values()])
                g.remove_node(gnodes[maxind])
            elif method == 'keepmin':
                minind = np.argmin([i['score'] for i in g.nodes.values()])
                neighbors = list(g.neighbors(gnodes[minind]))
                for n in neighbors:
                    g.remove_node(n)
            results += greduce(subgraphs(g), method=method)

    return results


def prune_matching_table(table, score, method='keepmin'):
    """eliminates match conflicts by comparing scores

    Parameters
    ----------
    table : list
        list (len = n_match_candidates) of numpy arrays (size = n_experiment)
        containing either cellIDs or -1
    score : :class:`numpy.ndarray`
        size = n_match_candidates, a float score value for each candidat
    method : str
        'keepmin' reduce graph by finding the minimum score of a subraph
            and eliminating its neighbors
        'popmax' reduce graph by finding the maximum score and eliminating it

    Returns
    -------
    pruned : list
        pruned version of the input table

    """
    G = nx.Graph()
    for i in range(len(table)):
        G.add_node(i, score=score[i])

    # if any 2 lines share an entry, make an edge between them
    for (i0, line0), (i1, line1) in itertools.combinations(
            enumerate(table), 2):
        nz = (line0 != -1) & (line1 != -1)
        if np.any(line0[nz] == line1[nz]):
            G.add_edge(i0, i1)

    inds = greduce(subgraphs(G), method=method)
    pruned = [table[i] for i in inds]

    return pruned


def pair_match_job(pair_args):
    pair_match = PairwiseMatching(input_data=pair_args, args=[])
    pair_match.run()
    delattr(pair_match, 'logger')
    return pair_match


class NwayMatching(ArgSchemaParser):
    default_schema = NwayMatchingSchema
    default_output_schema = NwayMatchingOutputSchema

    def make_masks_from_dicts(self):
        """read the input json, populate the experiments
        with nice masks.

        Parameters
        ----------
        input_json : str
            path to json file containing nway matching input

        """
        self.experiments = []
        for exp in self.args['experiment_containers']['ophys_experiments']:
            self.experiments.append(
                    utils.create_nice_mask(
                        exp,
                        self.args['output_directory']))

        if len(self.experiments) < 2:
            raise NwayException("Need at least 2 experiements from input")

    def gen_nway_table_with_redundancy(self):
        """Combine all pairwise matches into one table
        regardless of conflicts.

        """
        matching_frame = pd.DataFrame(
            columns=[e['id'] for e in self.experiments])

        for pair in self.pair_matches:
            with open(pair.args['output_json'], 'r') as f:
                pairj = json.load(f)
            pairframe = pd.DataFrame(
                    [[i['fixed'], i['moving']] for i in pairj['matches']],
                    columns=[
                        pairj['fixed_experiment'],
                        pairj['moving_experiment']])
            matching_frame = matching_frame.append(pairframe)

        return matching_frame

    @staticmethod
    def remove_nway_table_redundancy(frame):
        """Remove any matching lines or subsets in a table

        Parameters
        ----------
        frame : :class:`pandas.DataFrame` or :class:`numpy.ndarray`
            table of cellIDs or zeros

        Returns
        -------
        table_new : list
            list of lines without redundancy or subsets

        """

        table = np.array(frame).astype('float')
        table[np.isnan(table)] = -1
        table = table.astype('int')

        linenum, expnum = np.shape(table)
        stoptag = 0

        while stoptag == 0:
            table_new = [table[0]]

            for i, oline in enumerate(table[1:], 1):
                for j, nline in enumerate(table_new):
                    mergetag = 1
                    omiss = oline == -1
                    nmiss = nline == -1

                    # matches not in any of the same experiments
                    if np.all(omiss | nmiss):
                        mergetag = 0

                    # matches in the same experiment(s), but no match
                    if np.any(
                            (nline != oline) &
                            np.invert(omiss) &
                            np.invert(nmiss)):
                        mergetag = 0

                    if mergetag == 1:
                        for k in range(expnum):
                            if oline[k] != -1:
                                nline[k] = oline[k]
                        break
                        # break for loop j
                        # only merge to one previous record

                # old line unrelated to any new lines, add it
                if mergetag == 0:
                    table_new.append(oline)

            linenum_new = np.shape(table_new)[0]

            if linenum == linenum_new:
                stoptag = 1
            else:
                table = table_new
                linenum = linenum_new

        return table_new

    def prune_matching_graph(self, matching_table_nway):
        """remove matching conflicts by scoring match candidates
        and pruning according to that score

        Parameters
        ----------
        matching_table_nway : list
            list of match arrays

        Returns
        -------
        matching_table_nway_new : list
            pruned list of match arrays

        """

        nline = np.shape(matching_table_nway)[0]

        exids = [exp['id'] for exp in self.experiments]

        # NOTE
        # assign a score to each line in the matching table
        # the score is the sum of costs for each pair in the
        # line. The costs are generally in [0, 2] when the
        # pair was within max_distance and 1000 when the pair
        # was not.
        # the scores are used to choose between conflicting lines
        # for n experiments, there are np = (n)(n - 1) / 2 pairs
        # example 1:
        # line1 = [a, b, -1, -1], line2 = [a, c, -1, -1]
        # score1 = cost(a, b) + 5000, score2 = cost(a, c) + 5000
        # if cost(a, b) < cost(a, c), line1 is preferred
        # example 2:
        # line1 = [a, b, c, -1], line2 = [a, c, -1, -1]
        # score1 = cost(a, b) + cost(b, c) + cost(a, c) + 3000
        # score2 = cost(a, c) + 5000
        # score1 in range [3000, 3006], score2 in range [5000, 5002]
        # line1 is preferred

        score = np.zeros(nline)
        for pair in self.pair_matches:
            with open(pair.args['output_json'], 'r') as f:
                j = json.load(f)
            # all possible candidates within max_distance
            candidates = {
                    (m['fixed'], m['moving']): m['cost']
                    for k in ['matches', 'rejected']
                    for m in j[k]}
            prow = exids.index(pair.args['fixed']['id'])
            pcol = exids.index(pair.args['moving']['id'])
            for i in range(nline):
                col = matching_table_nway[i][pcol]
                row = matching_table_nway[i][prow]
                if (row, col) in candidates:
                    pscore = candidates[(row, col)]
                else:
                    pscore = 1000
                score[i] += pscore

        matching_table_nway_new = prune_matching_table(
                matching_table_nway, score, self.args['pruning_method'])

        return matching_table_nway_new

    def add_remaining_cells(self, matching_table_nway):
        """add any remaining cells that had no match

        Parameters
        ----------
        matching_table_nway : list
            list of match arrays

        Returns
        -------
        matching_table_nway_new : list
            with added unmatched cells

        """

        linenum = np.shape(matching_table_nway)[0]
        cnt = linenum
        matching_table_nway_new = matching_table_nway
        label_remain = []

        for j in range(len(self.experiments)):
            labels = [v['id'] for v in self.experiments[j]['cell_rois']]
            for i in range(linenum):
                labels = np.setdiff1d(labels, [matching_table_nway[i][j]])
            label_remain.append(labels)

        for i in range(len(self.experiments)):
            num = len(label_remain[i])
            for j in range(num):
                cnt = cnt + 1
                this_record = np.zeros(len(self.experiments)) - 1
                this_record[i] = label_remain[i][j]
                matching_table_nway_new.append(this_record.astype(int))

        return matching_table_nway_new

    def create_output_dict(self):
        """write the nway matching results to a dict

        Returns
        -------
        matchingdata: dict
           "nway_matches": list of sets of cellID matches
           "pairwise_results": list of results from the pairwise matching
        """
        cellnum = np.shape(self.matching_table_nway)[0]
        matchingdata = dict()
        matchingdata["nway_matches"] = []
        prob = np.zeros(len(self.experiments))

        for i in range(cellnum):
            matching_exp_num = \
                    len(self.experiments) - 1 - \
                    np.count_nonzero(self.matching_table_nway[i] == -1)
            prob[matching_exp_num] = prob[matching_exp_num] + 1
            self.logger.debug(self.matching_table_nway[i])

        for i in range(cellnum):
            thisrgn = [v for v in self.matching_table_nway[i] if v != -1]

            if len(thisrgn) > 0:
                matchingdata["nway_matches"].append(thisrgn)

        matchingdata["pairwise_results"] = []
        for k in self.pair_matches:
            with open(k.args['output_json'], 'r') as f:
                j = json.load(f)
            matchingdata['pairwise_results'].append(j)
            if not self.args['save_pairwise_results']:
                os.remove(k.args['output_json'])

        return matchingdata

    def run(self):
        """Nway cell matching by calling pairwise
           matching and then combining the results
        """
        # log this ENV variable, if present
        commit = os.environ.get("NWAY_COMMIT_SHA", None)
        self.logger.name = type(self).__name__
        self.logger.info(f"NWAY_COMMIT_SHA {commit}")
        self.logger.info(f"Nway matching version {nway_version}")

        self.make_masks_from_dicts()

        # pair-wise matching
        self.pair_matches = []
        pair_arg_list = []
        for fixed, moving in itertools.combinations(self.experiments, 2):
            pair_args = copy.deepcopy(self.args)
            # marshmallow 3.0.0rc6 is less forgiving about extra keys around
            # so, pop out the unused shared keys here
            for popkey in [
                    "pruning_method",
                    "experiment_containers",
                    "save_pairwise_results",
                    "parallel_workers"]:
                pair_args.pop(popkey)
            pair_args["fixed"] = fixed
            pair_args["moving"] = moving
            pair_args["output_json"] = os.path.join(
                    self.args["output_directory"],
                    "{}_to_{}_output.json".format(moving['id'], fixed['id']))
            pair_arg_list.append(pair_args)

        if self.args['parallel_workers'] == 1:
            self.pair_matches = [pair_match_job(i) for i in pair_arg_list]
        else:
            with multiprocessing.Pool(self.args['parallel_workers']) as pool:
                self.pair_matches = pool.map(pair_match_job, pair_arg_list)

        self.logger.name = type(self).__name__
        df, id_map = utils.summarize_registration_success(
                [p.args['output_json'] for p in self.pair_matches])
        self.logger.info(f"registration success(1) or failure (0):\n{df}\n"
                         f"id map{json.dumps(id_map, indent=2)}")
        if not df.all(axis=None):
            raise NwayException("not all pairwise registrations succeeded.")

        # generate N-way matching table
        matching_frame = self.gen_nway_table_with_redundancy()
        matching_table_nway_tmp = self.remove_nway_table_redundancy(
                matching_frame)
        matching_table_nway_tmp = self.prune_matching_graph(
                matching_table_nway_tmp)
        self.matching_table_nway = self.add_remaining_cells(
                matching_table_nway_tmp)
        self.logger.info("Nway matching is done!")

        # Generate first part of output dict (necessary for summary plots)
        output_dict = self.create_output_dict()

        # Generate nway match summary plots
        self.logger.info("Creating match summary plots")
        nway_summarizer = NwaySummary(
            input_data={'nway_input': self.args,
                        'nway_output': output_dict,
                        'output_directory': self.args["output_directory"]},
            args=[])
        plot_save_paths: dict = nway_summarizer.run()

        output_dict.update(plot_save_paths)
        self.output(output_dict, indent=2)
        self.logger.info("wrote {}".format(self.args['output_json']))


if __name__ == "__main__":  # pragma: no cover
    nmod = NwayMatching()
    nmod.run()



import cv2
import numpy as np
import logging
from typing import List, Tuple
import nway.image_processing_utils as impu

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class TransformException(Exception):
    pass


class Transform():
    """identity transform and base class for other transforms.

    Attributes
    ----------
    matrix: np.ndarray
        3 x 3 transformation matrix

    Methods
    -------
    estimate(src, dst):
        estimates the transform matrix. For generality, all transforms
        accept src and dst for estimate method, even when unused, as here
    apply(src, dst):
        applies the transform to src (moving image). For generality, both
        src and dst are potentially modified by apply()

    """
    def __init__(self):
        return

    def estimate(self, src: np.ndarray, dst: np.ndarray):
        """estimates transform matrix given src and dst.
        In this case, just identity.

        Parameters
        ----------
        src: np.ndarray
            uint8, needed for opencv operations
        dst: np.ndarray
            uint8, needed for opencv operations

        """
        self.matrix = np.eye(3).astype('float32')

    def apply(self, src: np.ndarray, dst: np.ndarray
              ) -> Tuple[np.ndarray, np.ndarray]:
        """applies a Euclidean transform to src with self.matrix

        Parameters
        ----------
        src: np.ndarray
            uint8, needed for opencv operations
        dst: np.ndarray
            uint8, needed for opencv operations

        Returns
        -------
        mov: np.ndarray
            uint8, transformed src image
        dst: np.ndarray
            uint8, transformed dst image. in this case, not transformed

        """
        mov = impu.warp_image(src, self.matrix, 'MOTION_EUCLIDEAN', dst.shape)
        return mov, dst


class CLAHE(Transform):
    """applies contrast limited adaptive histogram equalization
    to both src and dst images. Has no geometrical transform, yet
    inherits from Transform() to be easily swappable in a transform List.
    """
    def __init__(self, CLAHE_grid: int, CLAHE_clip: float):
        """
        Parameters
        ----------
        CLAHE_grid : int
            passed as tileGridSize to cv2.createCLAHE
        CLAHE_clip : float
            passed as clipLimit to cv2.createCLAHE
        """
        self.CLAHE_grid = CLAHE_grid
        self.CLAHE_clip = CLAHE_clip

    def apply(self, src: np.ndarray, dst: np.ndarray
              ) -> Tuple[np.ndarray, np.ndarray]:
        """applies CLAHE to both src and dst

        Parameters
        ----------
        src: np.ndarray
            uint8, needed for opencv operations
        dst: np.ndarray
            uint8, needed for opencv operations

        Returns
        -------
        mov: np.ndarray
            uint8, contrast-adjusted src image
        dst: np.ndarray
            uint8, contrast-adjusted dst image.

        """
        mov = impu.contrast_adjust(src, self.CLAHE_grid, self.CLAHE_clip)
        fix = impu.contrast_adjust(dst, self.CLAHE_grid, self.CLAHE_clip)
        return mov, fix


class PhaseCorrelate(Transform):
    """estimates geometrical transform by phase correlation
    and applies the transform
    """
    def estimate(self, src: np.ndarray, dst: np.ndarray):
        """
        Parameters
        ----------
        src: np.ndarray
            uint8, needed for opencv operations
        dst: np.ndarray
            uint8, needed for opencv operations

        """
        (dx, dy), _ = cv2.phaseCorrelate(src1=dst.astype('float32'),
                                         src2=src.astype('float32'))
        self.matrix = np.eye(3).astype('float32')
        self.matrix[0, 2] = dx
        self.matrix[1, 2] = dy

    def apply(self, src: np.ndarray, dst: np.ndarray
              ) -> Tuple[np.ndarray, np.ndarray]:
        mov = impu.warp_image(src, self.matrix, 'MOTION_EUCLIDEAN', dst.shape)
        return mov, dst


class Crop(Transform):
    """crops both src and dst image, default identity transform
    """
    def __init__(self, edge_buffer: int):
        """

        Parameters
        ----------
        edge_buffer: int
            this many pixels will be cropped from top, bottom, left, and right

        """
        self.edge_buffer = edge_buffer

    def apply(self, src: np.ndarray, dst: np.ndarray
              ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Parameters
        ----------
        src: np.ndarray
            uint8, needed for opencv operations
        dst: np.ndarray
            uint8, needed for opencv operations

        Returns
        -------
        cropped_src: np.ndarray
            uint8, cropped src image
        cropped_dst: np.ndarray
            uint8, cropped dst image.

        """
        if self.edge_buffer == 0:
            return src, dst
        if np.any(self.edge_buffer >
                  0.5 * np.array((src.shape, dst.shape)).flatten()):
            raise TransformException(f"edge buffer of {self.edge_buffer} "
                                     "is too large for images of shapes "
                                     f"{src.shape}, {dst.shape}")
        cropped_src = np.copy(src[
            self.edge_buffer:-self.edge_buffer,
            self.edge_buffer:-self.edge_buffer])
        cropped_dst = np.copy(dst[
            self.edge_buffer:-self.edge_buffer,
            self.edge_buffer:-self.edge_buffer])
        return cropped_src, cropped_dst


class ECC(Transform):
    """estimates a transform between two images with cv2.findTransformECC
    and applies the transform to one of the images.
    """
    def __init__(self, maxCount: int, epsilon: float,
                 motion_type: str, gaussFiltSize: int):
        self.maxCount = maxCount
        self.epsilon = epsilon
        self.gaussFiltSize = gaussFiltSize
        self.motion_type = motion_type

    @classmethod
    def cv2_motion(self, motion_type: str) -> int:
        cvmotion = {
                "MOTION_TRANSLATION": cv2.MOTION_TRANSLATION,
                "MOTION_EUCLIDEAN": cv2.MOTION_EUCLIDEAN,
                "MOTION_AFFINE": cv2.MOTION_AFFINE,
                "MOTION_HOMOGRAPHY": cv2.MOTION_HOMOGRAPHY}
        return cvmotion[motion_type]

    def estimate(self, src: np.ndarray, dst: np.ndarray):
        """estimates a transform from src to dst with
        cv2.findTransformECC
        Parameters
        ----------
        src: np.ndarray
            uint8, needed for opencv operations
        dst: np.ndarray
            uint8, needed for opencv operations

        """
        criteria = (cv2.TERM_CRITERIA_EPS | cv2.TERM_CRITERIA_COUNT,
                    self.maxCount, self.epsilon)
        tform = np.array([[1.0, 0.0, 0.0],
                          [0.0, 1.0, 0.0]]).astype('float32')
        if self.motion_type == 'MOTION_HOMOGRAPHY':
            hrow = np.array([0.0, 0.0, 1.0]).astype('float32')
            tform = np.vstack((tform, hrow))
        try:
            # Run the ECC algorithm. The results are stored in tform
            ccval, tform = cv2.findTransformECC(
                    templateImage=dst,
                    inputImage=src,
                    warpMatrix=tform,
                    motionType=self.cv2_motion(self.motion_type),
                    criteria=criteria,
                    inputMask=None,
                    gaussFiltSize=self.gaussFiltSize)
        except cv2.error as cv2err:
            err = TransformException("findTransformECC failed to align images")
            err.args = err.args + cv2err.args
            raise err

        if tform.shape == (2, 3):
            tform = np.vstack((tform, [0, 0, 1]))
        self.matrix = tform

    def apply(self, src: np.ndarray, dst: np.ndarray
              ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Parameters
        ----------
        src: np.ndarray
            uint8, needed for opencv operations
        dst: np.ndarray
            uint8, needed for opencv operations

        Returns
        -------
        mov: np.ndarray
            uint8, transformed src image
        dst: np.ndarray
            original dst image

        """
        mov = impu.warp_image(src, self.matrix, self.motion_type, dst.shape)
        return mov, dst


class TransformList():
    """a list of Transforms that behaves like a Transform
    """
    def __init__(self, transforms: List[Transform]):
        """

        Parameters
        ----------
        transforms: List[Transform]
            List of Transform instances, or Transform-based instances

        """
        self.transforms = transforms

    def estimate(self, src, dst):
        """estimates the individual transforms by successively
        estimating and applying them.

        Parameters
        ----------
        src: np.ndarray
            uint8, needed for opencv operations
        dst: np.ndarray
            uint8, needed for opencv operations

        """
        mov = np.copy(src)
        fix = np.copy(dst)
        for transform in self.transforms:
            transform.estimate(mov, fix)
            mov, fix = transform.apply(mov, fix)

    def apply(self, src, dst
              ) -> Tuple[np.ndarray, np.ndarray]:
        """successively applies transforms to a pair of images

        Parameters
        ----------
        src: np.ndarray
            uint8, needed for opencv operations
        dst: np.ndarray
            uint8, needed for opencv operations

        Returns
        -------
        mov: np.ndarray
            uint8, transformed src image
        fix: np.ndarray
            uint8, transformed dst image

        """
        mov = np.copy(src)
        fix = np.copy(dst)
        for transform in self.transforms:
            mov, fix = transform.apply(mov, fix)
        return mov, fix

    @property
    def matrix(self) -> np.ndarray:
        """the cumulative geometric transform matrix
        for the list of transforms

        Returns
        -------
        tform: np.ndarray
            3x3 float32 transform matrix
        """
        tform = np.eye(3).astype('float32')
        for transform in self.transforms:
            tform = transform.matrix.dot(tform)
        return tform
